/********************************************************************
Copyright (c) 2010-2015, Regents of the University of Colorado

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of the University of Colorado nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
********************************************************************/

#include "Quip.h"
#include <queue>
#include <sstream>
#include <fstream>
#include <boost/algorithm/string.hpp>
#include <algorithm>
#include <boost/functional/hash.hpp>
#include <chrono>

using namespace std;

namespace std {
  std::size_t hash<Quip::Cube>::operator()(const Quip::Cube& c) const
  {
    using boost::hash_range;

    std::size_t seed = 0;

    hash_range(seed, c.begin(), c.end());

    return seed;
  }
}
  
namespace Quip {
  // Should be a map so references remain valid across operations
  // (not e.g., unordered_map)
  typedef std::map<CubeID, Lemma> LemmaDatabase;

  // This allows the cleaner logging interface while still avoiding expensive
  // calls to stringOfClauseOrCube when the string would be sent to a
  // null_stream_buf. Two alternatives are: 1) having every log output wrapped
  // in if(verbosity > ...); or 2) using calls like logger.verbose() <<
  // stringOfClauseOrCube(...). (1) creates lots of extra if conditions, (2)
  // makes expensive calls to stringOfClauseOrCube when verbosity is low
  // (profiling showed it was ~15% of runtime).  Using this, we pass a
  // PrintCubePair, and if the stream is a null_stream_buf, it never needs to
  // do the expensive operations.
  std::ostream& operator<<(std::ostream& os, PrintCubePair cube_ev)
  {
    if (!os.good()) {
      return os;
    }
  
    Expr::Manager::View * ev;
    // Could be a clause too
    const Cube * cube;
    std::tie(cube, ev) = cube_ev;
    assert(cube);
    assert(ev);
  
    for (auto it = cube->begin(); it != cube->end(); ++it) {
      std::string strid = Expr::stringOf(*ev, *it);
      os << " " << strid;
    }
  
    return os;
  }

  QuipLog null_log;

  // Instrumentation-related classes
  class QuipTimer {
    std::chrono::steady_clock::time_point begin_time, end_time;

    public:
      void begin() {
        begin_time = std::chrono::steady_clock::now();
      }

      void end() {
        end_time = std::chrono::steady_clock::now();
      }

      double elapsed() {
        std::chrono::duration<double> elapsed = 
          std::chrono::duration_cast<std::chrono::duration<double> >
          (end_time - begin_time);
        return elapsed.count();
      }
  };

  // Times the block in which it is instantiated, adds the elapsed time
  // to target
  class QuipAutoTimer {
    QuipTimer timer;
    double & target;

    public:
      QuipAutoTimer(double & target) : target(target) 
      {
        timer.begin();
      }

      ~QuipAutoTimer() {
        timer.end();
        target += timer.elapsed();
      }
  };

  void primeIDVector(Expr::Manager::View * ev,  Expr::IDVector & vec) {
    std::transform(vec.begin(), vec.end(), vec.begin(), [&](ID x){ return ev->prime(x); });
  }

  // Configuration for support graph types - should definitely be considered
  // obsolete as EXPERIMENTAL_SG (i.e., TRUSS) is the only thing that really
  // works well
  bool isCoreBasedSG(SupportGraphType type) {
    switch(type) {
        case NO_SG:
        case ACTIVATION_LITERAL_SG:
        case FINAL_ONLY_SG:
        case EXPERIMENTAL_SG:
        case HYBRID_SG:
          return false;
        case UNSAT_CORE_SG:
          return true;
        default:
          assert(false);
      }
  }

  bool isLitBasedSG(SupportGraphType type) {
    switch(type) {
        case ACTIVATION_LITERAL_SG:
        case FINAL_ONLY_SG:
        case EXPERIMENTAL_SG:
        case HYBRID_SG:
        case NO_SG:
          assert(!isCoreBasedSG(type));
          return true;
        case UNSAT_CORE_SG:
          return false;
        default:
          assert(false);
      }
  }

  bool requiresSGOnPush(SupportGraphType type) {
    switch(type) {
        case ACTIVATION_LITERAL_SG:
        case HYBRID_SG:
        case UNSAT_CORE_SG:
          return true;
        case NO_SG:
        case FINAL_ONLY_SG:
        case EXPERIMENTAL_SG:
          return false;
        default:
          assert(false);
    } 
  }

  bool requiresSGOnConsecution(SupportGraphType type) {
    switch(type) {
        case ACTIVATION_LITERAL_SG:
          return true;
        case UNSAT_CORE_SG:
        case NO_SG:
        case FINAL_ONLY_SG:
        case EXPERIMENTAL_SG:
        case HYBRID_SG:
          return false;
        default:
          assert(false);
    }
  }

  bool forbidsSGOnConsecution(SupportGraphType type) {
    switch(type) {
        case NO_SG:
        case FINAL_ONLY_SG:
          return true;
        case UNSAT_CORE_SG:
        case ACTIVATION_LITERAL_SG:
        case EXPERIMENTAL_SG:
        case HYBRID_SG:
          return false;
        default:
          assert(false);
    }
  }

  // std::list is guaranteed not to invalidate pointers, and we maintain
  // many pointers into the pool of obligations
  typedef std::list<ProofObligation> ObligationPool;
  typedef priority_queue<ProofObligation*, vector<ProofObligation*>, ProofObligationComparator > ObligationQueue;

  ProofObligation * newObligation(ObligationPool & p,
                                  const Cube & cube, 
                                  int level, 
                                  ProofKind type, 
                                  ProofObligation * parent = NULL, 
                                  int effort_limit=INT_MAX) {
    p.push_back(ProofObligation(cube, level, type, parent, effort_limit));
    return &p.back();
  }
  
  ProofObligation * newObligation(ObligationPool & p,
                                  const ProofObligation & other) {
    p.push_back(ProofObligation(other));
    return &p.back();
  }

  /***
  Sets the proof obligation's active to false and propagates the effort value
  upwards until reaching an active parent. 
  ***/
  void deactivateProofObligation(ProofObligation * po) {
    po->active = false;
    while (po->parent) {
      po->parent->effort += po->effort;
      if (po->parent->active) break;
      po = po->parent;
    }
  }
  
  /***
  Create and add a new reachable state.
  The state is added in the location defined by the manager's insertion_policy
  ***/
  ReachableState * ReachableStateManager::addState(const Cube & cube, const Cube & inputs)
  {
    //Check if the new state already exists in the list. 
    for (list<ReachableState*>::iterator it=states.begin(); it!=states.end(); it++){
      if ((*it)->state == cube){
        return *it;
      }
    }
      
    total_states++;
    ReachableState * new_state = new ReachableState(cube, inputs);
    //find pointer to the position at which to insert 
    int pos = states.size() * (1-insertion_policy);
    list<ReachableState*>::iterator it=states.end();
    for (; it!=states.begin(); it--){
      if (pos == 0){
        break;
      }
      pos--;
    }
    states.insert(it, new_state);
   
    //evict last state
    if ((int)states.size() > maxn_states){
      num_evictions++;
      list<ReachableState*>::iterator last = --states.end();
      evicted_states.push_back(*last);
      states.erase(last);
    }
    return new_state;
  }
  
  /***
  Return first (most active) reachable state which intersects with cube, or
  NULL if no state intersects with cube.
  Automatically promote the intersecting state to the front of the list.
  ***/
  ReachableState * ReachableStateManager::intersect(const Cube & cube)
  {
    /*Need to check every state explicitly - cannot use set::find or set 
    binary searching because cube may not be EQUAL to any cubes in reachable - 
    it just needs to be a subset of it. Also the cube ordering is meaningless since 
    it could be any subset of literals.*/
    for (list<ReachableState*>::iterator it = states.begin(); it != states.end(); it++){
      if (QuipSolver::subsumes((*it)->state, cube)) {
        ReachableState * ret = *it;
        //move ret to the front of the list
        states.erase(it);
        states.push_front(ret);
        num_intersections++;
        return ret;
      }
    }
    return NULL;
  }

  const CubeID CUBE_ID_NULL = 0;  

  /***
  Database to manage lemmas. The Lemma structure stores the metadata
  about a lemma. This CubeDatabase handles access to the data. 
  ***/
  class CubeDatabase {
    std::unordered_map<Cube, CubeID> cube_to_id;
    std::unordered_map<CubeID, Cube> id_to_cube;
    CubeID last_id;
    LemmaDatabase lemma_db;

    CubeID getNextID() {
      last_id++;
      return last_id;
    }

    void addCubeWithID(const Cube & cube, CubeID id, const Lemma & lem) {
      assert(cube_to_id.find(cube) == cube_to_id.end());
      assert(id_to_cube.find(id) == id_to_cube.end());
      cube_to_id[cube] = id;
      id_to_cube[id] = cube;

      assert(lemma_db.find(id) == lemma_db.end());

      lemma_db[id] = lem;
    }

    public:
      CubeDatabase() : last_id(CUBE_ID_NULL) { }
      bool isKnown(const Cube & cube) const;
      bool isKnown(CubeID id) const;
      CubeID addLemma(const Lemma & lem);
      CubeID getID(const Cube & cube) const; 
      const Cube & getCube(CubeID id) const;
      Lemma & getLemma(const Cube & cube);
      Lemma & getLemma(CubeID cube);
      const Lemma & getLemma(const Cube & cube) const;
      const Lemma & getLemma(CubeID cube) const;
  };
    
  typedef std::set<CubeID> SupportSet;

  class SupportGraph {
    const CubeDatabase & cube_db;
    QuipLog & logger; 
    typedef std::unordered_map<CubeID, SupportSet> SupportMap;
    SupportMap cube_in_edges;
    SupportMap cube_out_edges;
    std::set<CubeID> supported_cubes;
    std::set<CubeID> known_cubes;

    bool isBetter(const SupportSet & lhs, const SupportSet & rhs) const;
    unsigned countLits(const SupportSet & ss) const;
    
    //stats    
    int total, total_cnt;
    vector<int> in_deg_dist, out_deg_dist;
    int needed_inits;
    int unknown_supp;

    void findUnsupportedCubes(const Cube & init_cube, std::set<CubeID> & unsupported) const;
    
    void addEdge(CubeID from, CubeID to);
    void removeEdge(CubeID from, CubeID to);
    bool removeEdgeIfExists(CubeID from, CubeID to);

    public:
      SupportGraph(CubeDatabase & db) : cube_db(db), logger(null_log) { }
      SupportGraph(CubeDatabase & db, QuipLog & logger) : cube_db(db), logger(logger) { }

      // Delete a cube - i.e., anything that it supports must have its support
      // set deleted. We do not delete the cube in any other sense
      void deleteCube(const Cube & cube);
      void deleteCube(const CubeID id);
      
      void recordSupport(const set<Cube> & support_set, const Cube & supported);
      void recordSupport(const set<CubeID> & support_set, CubeID & supported);

      void deleteSupport(const Cube & supported);
      void deleteSupport(CubeID & supported);

      std::pair<SupportSet::const_iterator, SupportSet::const_iterator>
        getInEdges(const Cube & cube) const;
      std::pair<SupportSet::const_iterator, SupportSet::const_iterator>
        getInEdges(CubeID id) const;
      std::pair<SupportSet::const_iterator, SupportSet::const_iterator>
        getOutEdges(const Cube & cube) const;
      std::pair<SupportSet::const_iterator, SupportSet::const_iterator>
        getOutEdges(CubeID id) const;
      std::pair<std::set<CubeID>::const_iterator, std::set<CubeID>::const_iterator> 
        getSupportedCubeIDs() const;
      std::pair<std::set<CubeID>::const_iterator, std::set<CubeID>::const_iterator> 
        getKnownCubeIDs() const;
      std::set<Cube> getSupportSet(const Cube & cube) const;

      bool hasSupportSet(const Cube & cube) const;
      bool hasSupportSet(CubeID id) const;

      size_t getInDegree(const Cube & cube) const;
      size_t getInDegree(CubeID id) const;
      size_t getOutDegree(const Cube & cube) const;
      size_t getOutDegree(CubeID id) const;

      int getLevel(const Cube & cube) const;
      int getLevel(CubeID id) const;
      
      int getMinSupportLevel(const Cube & cube) const;
      int getMinSupportLevel(CubeID id) const;

      bool isBad(const Cube & cube) const;
      bool isBad(CubeID id) const;

      // Builds an inductive trace containing all cubes supported relative
      // to the given init_cube. Assumes F_inf as known to this support
      // graph is still correct.
      void buildInductiveTrace(const Cube & init_cube, vector<CubeSet> & frames) const;

      std::string getAnnotation(const Cube & cube) const;
      std::string getAnnotation(CubeID cube) const;
      bool hasAnnotation(const Cube & cube) const;
      bool hasAnnotation(CubeID cube) const;

      void writeInfDot(std::ostream & os, const Cube & property = Cube()) const;
      void writeInfDot(const std::string & file, const Cube & property = Cube()) const;
      void writeDot(std::ostream & os, const Cube & property = Cube()) const;
      void writeDot(const std::string & file, const Cube & property = Cube()) const;
      
      void computeStats(const Cube & property, bool inf_only=false);
      void printStats(bool inf_only=false);
  };

  class SupportGraphDotWriter {
    typedef std::unordered_map<std::string, std::string> ParamMap;
    const SupportGraph & graph;
    CubeID property_id;

    std::set<CubeID> proof_cubes;
    std::unordered_map<CubeID, unsigned> cube_to_outdegree;
    // Cubes at index i are from level i
    // It's a map because we can have -1 (we weren't given a level for the 
    // cube, though maybe this should be classed as a bug) or INT_MAX
    // (level infinity)
    std::map< int, std::set<CubeID> > ranked_cubes;

    std::unordered_map<CubeID, unsigned> computeOutDegrees() const;
    std::map<int, std::set<CubeID> > computeRanks() const;

    std::string dotPreamble() const;
    std::string dotPostamble() const;
    std::string dotCubeIdentifier(CubeID cube) const;
    std::string dotCube(CubeID cube) const;
    std::string dotEdge(CubeID from, CubeID to) const;
    std::string dotParams(const ParamMap & params) const;

    public:
    SupportGraphDotWriter(const SupportGraph & graph, CubeID property_id) : 
      graph(graph),
      property_id(property_id) 
    { 
      proof_cubes = findTransitiveSupport(property_id);
      cube_to_outdegree = computeOutDegrees();
      ranked_cubes = computeRanks();
    }

    void writeInfDot(std::ostream & os) const;
    void writeDot(std::ostream & os) const;
    std::unordered_map<CubeID, unsigned>& getOutDegrees();
    void findTransitiveSupport(std::set<CubeID> & support) const;
    std::set<CubeID> findTransitiveSupport(CubeID cube) const;
  };

  std::map<int, std::set<CubeID> > SupportGraphDotWriter::computeRanks() const {
    std::map<int, std::set<CubeID> > ranks; 
    auto cube_range = graph.getKnownCubeIDs();
    for(auto it = cube_range.first; it != cube_range.second; ++it) {
      CubeID id = *it;
      if(id != property_id) { 
        int level = graph.getLevel(id);
        // Leave nodes for which we don't have a level free
        if(level >= 0) {
          ranks[level].insert(id);
        }
      }
    }

    return ranks;
  }

  std::unordered_map<CubeID, unsigned > SupportGraphDotWriter::computeOutDegrees() const {
    std::unordered_map<CubeID, unsigned > cube_to_outdegree;
    auto known_cube_range = graph.getKnownCubeIDs();
    for(auto it = known_cube_range.first; it != known_cube_range.second; ++it) {
      CubeID cube = *it;
      cube_to_outdegree[cube] = 0;
    }

    auto supp_cube_range = graph.getSupportedCubeIDs();
    for(auto it = supp_cube_range.first; it != supp_cube_range.second; ++it) {
      CubeID supported = *it;
      auto support_range = graph.getInEdges(supported);

      for(auto cube_it = support_range.first; cube_it != support_range.second; ++cube_it) {
        CubeID cube = *cube_it;
        cube_to_outdegree[cube]++;
      }
    }
    return cube_to_outdegree;
  }
  
  std::unordered_map<CubeID, unsigned>& SupportGraphDotWriter::getOutDegrees() {
    return cube_to_outdegree;
  }

  std::set<CubeID> SupportGraphDotWriter::findTransitiveSupport(CubeID cube) const {
    std::set<CubeID> support;
    support.insert(cube);
    findTransitiveSupport(support);
    return support;
  }

  // Not as efficient as it could be
  void SupportGraphDotWriter::findTransitiveSupport(std::set<CubeID> & support) const {
    std::set<CubeID> support_updated = support;
    for(auto it = support.begin(); it != support.end(); ++it) {
      CubeID cube = *it;
      auto support_range = graph.getInEdges(cube);
      support_updated.insert(support_range.first, support_range.second);
    }

    if(support_updated.size() > support.size()) {
      support = support_updated;
      findTransitiveSupport(support);
    }
  }

  std::string SupportGraphDotWriter::dotPreamble() const {
    return "digraph support_graph {";
  }

  std::string SupportGraphDotWriter::dotPostamble() const {
    return "}";
  }

  std::string SupportGraphDotWriter::dotCubeIdentifier(CubeID cube) const {
    return "c_" + std::to_string(cube);
  }

  std::string SupportGraphDotWriter::dotParams(const ParamMap & params) const {
    if(params.empty()) {
      return "";
    }

    std::stringstream ss;

    ss << "[";
    for(auto pv : params) {
      std::string param = pv.first;
      std::string value = pv.second;

      // Dot tools seem to accept the extra commas, so we'll leave them
      // (e.g., [color=red,])
      ss << param << "=" << value << ",";
    }
    ss << "]";

    return ss.str();
  }

  std::string SupportGraphDotWriter::dotCube(CubeID cube) const {
    std::stringstream ss;
    bool necessary = proof_cubes.find(cube) != proof_cubes.end();

    std::string label = std::to_string(cube);

    unsigned in_degree = graph.getInDegree(cube);
    unsigned out_degree = cube_to_outdegree.at(cube);
    int level = graph.getLevel(cube);
    bool bad = graph.isBad(cube);

    std::string annotation = "";
    if(graph.hasAnnotation(cube)) {
      annotation = "[" + graph.getAnnotation(cube) + "]";
    }

    std::stringstream tooltip_ss;
    tooltip_ss << "\""
               << "F_" << (level == INT_MAX ? "inf" : std::to_string(level)) << " "
               << "in = " << in_degree << " "
               << "out = " << out_degree << " "
               << annotation
               << "\"";
    std::string tooltip = tooltip_ss.str();
 

    std::string color = "black";
    if(level == INT_MAX) {
      color = "orange";
    } else if(level == 0) {
      color = "grey";
    }
    if(necessary) {
      color = "red";
    }
    if(cube == property_id) {
      color = "\"#008B45\""; // Dark green 
    }

    if(bad) {
      assert(level < INT_MAX);
      assert(cube != property_id);
      color = "dodgerblue";
    }
   
    ParamMap p = {{ "label", "\"" + label + "\""}, 
                  { "fontcolor", color },
                  { "color", color },
                  { "tooltip", tooltip }};

    ss << dotCubeIdentifier(cube) 
       << dotParams(p)
       << ";"; 

    return ss.str();
  }

  std::string SupportGraphDotWriter::dotEdge(CubeID from, CubeID to) const {
    std::stringstream ss;

    bool from_necessary = proof_cubes.find(from) != proof_cubes.end();
    bool to_necessary = proof_cubes.find(to) != proof_cubes.end();
    std::string edgecolor = "black";
    if(to_necessary) {
      edgecolor = "red";
    } else if(from_necessary) {
      edgecolor = "orange";
    }

    ParamMap p = {{"color", edgecolor}};

    ss << dotCubeIdentifier(from) 
       << " -> " 
       << dotCubeIdentifier(to)
       << dotParams(p)
       << ";";

    return ss.str();
  }

  void SupportGraphDotWriter::writeInfDot(std::ostream & os) const {
    os << dotPreamble() << std::endl;

    auto cube_range = graph.getKnownCubeIDs();

    // Declare nodes
    for(auto it = cube_range.first; it != cube_range.second; ++it) {
      CubeID id = *it;
      if(graph.getLevel(id) == INT_MAX || id == property_id) {
        os << dotCube(id) << std::endl;
      }
    }

    // Output edges
    for(auto it = cube_range.first; it != cube_range.second; ++it) {
      CubeID id = *it;

      if(graph.getLevel(id) != INT_MAX && id != property_id) { continue; }
      auto support_range = graph.getInEdges(id);
      for(auto support_it = support_range.first; support_it != support_range.second; ++support_it) {
        CubeID support_id = *support_it;
        if(graph.getLevel(support_id) == INT_MAX || id == property_id) {
          os << dotEdge(support_id, id) << std::endl;;
        }
      }
    }

    os << dotPostamble() << std::endl;
  }

  void SupportGraphDotWriter::writeDot(std::ostream & os) const {
    os << dotPreamble() << std::endl;

    auto known_cube_range = graph.getKnownCubeIDs();
    auto supported_cube_range = graph.getSupportedCubeIDs();

    // Output edges
    // Suppress vertices for cubes in F_inf that have no out edges
    std::set<CubeID> suppressed_cubes;
    if(ranked_cubes.find(INT_MAX) != ranked_cubes.end()) {
     suppressed_cubes = ranked_cubes.at(INT_MAX);
    }
    suppressed_cubes.insert(property_id);

    std::stringstream edge_ss;
    for(auto it = supported_cube_range.first; it != supported_cube_range.second; ++it) {
      CubeID id = *it;

      if(id == property_id) {
        continue;
      }

      auto support_range = graph.getInEdges(id);
      bool id_inf = graph.getLevel(id) == INT_MAX;


      for(auto support_it = support_range.first; support_it != support_range.second; ++support_it) {
        CubeID support_id = *support_it;

        bool support_inf = graph.getLevel(support_id) == INT_MAX;

        // Don't put edges between infinity lemmas in, and also don't
        // draw edges involving the property
        if(id != property_id && (!id_inf || !support_inf)) {
          edge_ss << dotEdge(support_id, id) << std::endl;
          suppressed_cubes.erase(id);
          suppressed_cubes.erase(support_id);
        }
      }
    }

    // Declare nodes that we want to show
    for(auto it = known_cube_range.first; it != known_cube_range.second; ++it) {
      CubeID id = *it;
      if(suppressed_cubes.find(id) == suppressed_cubes.end()) {
        os << dotCube(id) << std::endl;
      }
    }

    // Arrange nodes into ranks by level. ranked_cubes should be an ordered
    // map for this to work
    // Lots of hacks in here
    for(auto it = ranked_cubes.begin(); it != ranked_cubes.end(); ++it) {
      int level = it->first;

      if(level == INT_MAX) {
        os << "[style=invis]";
      }

      os << std::endl;
      if(level == 0 || level == INT_MAX) {
        os << "{rank=source; ";
      } else {
        os << "{rank=same; ";
      }

      const std::set<CubeID> & cubes_at_rank = it->second;
      for(auto cube_it = cubes_at_rank.begin(); cube_it != cubes_at_rank.end(); ++cube_it) {
        CubeID id = *cube_it;
        assert(id != property_id);
        if(suppressed_cubes.find(id) == suppressed_cubes.end()) {
          os << dotCubeIdentifier(id) << " ";
        }
      }
      os << "}";
    }

    // Add one invisible edge between each pair of consecutive ranks, 
    // just in case there are not real ones

    ID last_id = CUBE_ID_NULL;
    for(auto it = ranked_cubes.begin(); it != ranked_cubes.end(); ++it) {
      const std::set<CubeID> & cubes_at_rank = it->second;
      if(!cubes_at_rank.empty()) {
        CubeID id = (*cubes_at_rank.begin());
        if(last_id != CUBE_ID_NULL) {
          os << dotCubeIdentifier(last_id) << " -> " << dotCubeIdentifier(id) << "[style=invis];" << std::endl;
        }
        last_id = id;
      }
    }

    os << edge_ss.str() << std::endl;
    
    os << dotPostamble() << std::endl;
  }

  /***
  Add an edge to the graph. It must not already exist.
  ***/
  void SupportGraph::addEdge(CubeID from, CubeID to) {
    assert(cube_db.isKnown(from));
    assert(cube_db.isKnown(to));

    assert(cube_in_edges[to].find(from) == cube_in_edges[to].end());
    assert(cube_out_edges[from].find(to) == cube_out_edges[from].end());

    cube_in_edges[to].insert(from);
    cube_out_edges[from].insert(to);

    known_cubes.insert(to);
    known_cubes.insert(from);
    supported_cubes.insert(to);
  }

  /***
  Remove an edge from the graph. The edge must exist.
  ***/
  void SupportGraph::removeEdge(CubeID from, CubeID to) {
    assert(removeEdgeIfExists(from, to));  
  }

  /***
  Remove an edge from the graph, returning true if and only if the
  edge was present to begin with
  ***/
  bool SupportGraph::removeEdgeIfExists(CubeID from, CubeID to) {
    assert(cube_db.isKnown(from));
    assert(cube_db.isKnown(to));

    SupportSet & in_edges = cube_in_edges[to];
    SupportSet & out_edges = cube_out_edges[from];

    auto in_iter = in_edges.find(from);
    auto out_iter = out_edges.find(to);

    bool in_present = in_iter != in_edges.end();
    bool out_present = out_iter != out_edges.end();

    // If the edge is only represented in one direction that's bad
    assert(in_present == out_present);

    bool present = in_present;
    if(present) {
      in_edges.erase(in_iter);
      out_edges.erase(out_iter);
      return true;
    } else {
      return false;
    }
  }

  unsigned SupportGraph::countLits(const SupportSet & ss) const {
    unsigned total = 0;
    for(CubeID id : ss) {
      assert(cube_db.isKnown(id));
      total += cube_db.getCube(id).size();
    }
    
    return total;
  }

  bool SupportGraph::isBetter(const SupportSet & lhs, const SupportSet & rhs) const {
    if(lhs == rhs) {
      return false;
    }

    if(lhs.size() < rhs.size()) { 
      return true;
    } else if (rhs.size() < lhs.size()) {
      return false;
    }

    // Size is equal

    size_t lhs_lits = countLits(lhs);
    size_t rhs_lits = countLits(rhs);

    if(lhs_lits < rhs_lits) {
      return true;
    } else if(rhs_lits < lhs_lits) {
      return false;
    }

    // Size and total literals is equal
    // Break tie arbitrarily based on Cube IDs
    auto lhs_it = lhs.begin();
    auto rhs_it = rhs.begin();

    while(lhs_it != lhs.end() && rhs_it != rhs.end()) {
      CubeID lhs_id = *lhs_it;
      CubeID rhs_id = *rhs_it;

      if(lhs_id < rhs_id) {
        return true;
      } else if(rhs_id < lhs_id) {
        return false;
      }
      
      lhs_it++;
      rhs_it++;
    }

    // We can only get here if they are the same, but we already
    // checked that at the beginning
    assert(false);
  }

  std::string SupportGraph::getAnnotation(const Cube & cube) const {
    assert(cube_db.isKnown(cube));
    CubeID id = cube_db.getID(cube);
    return getAnnotation(id);
  }

  std::string SupportGraph::getAnnotation(CubeID cube) const {
    assert(cube_db.isKnown(cube));
    return cube_db.getLemma(cube).annotation;
  }

  bool SupportGraph::hasAnnotation(const Cube & cube) const {
    assert(cube_db.isKnown(cube));
    CubeID id = cube_db.getID(cube);
    return hasAnnotation(id);
  }

  bool SupportGraph::hasAnnotation(CubeID cube) const {
    assert(cube_db.isKnown(cube));
    return cube_db.getLemma(cube).annotation != "";
  }

  void SupportGraph::deleteCube(const Cube & cube) {
    assert(cube_db.isKnown(cube));
    CubeID id = cube_db.getID(cube);
    deleteCube(id);
  }

  void SupportGraph::deleteCube(const CubeID id) {
    auto known_range = getKnownCubeIDs();
    for(auto it = known_range.first; it != known_range.second; ++it) {
      CubeID known_id = *it;
      if(hasSupportSet(known_id)) {
        auto support_range = getInEdges(known_id);
        std::set<CubeID> support_set(support_range.first, support_range.second);
        if(support_set.find(id) != support_set.end()) {
          cube_in_edges.erase(cube_in_edges.find(id));
        }
      }
    } 
  }
 
  void SupportGraph::recordSupport(const set<Cube> & support_set, const Cube & supported) {
    SupportSet ss;
    for(auto it = support_set.begin(); it != support_set.end(); ++it) {
      assert(cube_db.isKnown(*it));
      CubeID ss_id = cube_db.getID(*it);
      ss.insert(ss_id);
    }

    assert(cube_db.isKnown(supported));
    CubeID supp = cube_db.getID(supported);

    recordSupport(ss, supp);
  }

  void SupportGraph::recordSupport(const set<CubeID> & support_set, CubeID & supported) {
    auto support_it = cube_in_edges.find(supported);

    if(support_it == cube_in_edges.end()) { 
      // Just take the support set if we don't have one yet
      for(CubeID from_id : support_set) {
        addEdge(from_id, supported);
      }
    } else if(isBetter(support_set, support_it->second)) {
      deleteSupport(supported);

      for(CubeID from_id : support_set) {
        addEdge(from_id, supported);
      }
    }
  }

  /***
  Delete the support set of the cube. 
  ***/
  void SupportGraph::deleteSupport(const Cube & supported) {
    assert(cube_db.isKnown(supported));
    CubeID id = cube_db.getID(supported);
    deleteSupport(id);
  }

  void SupportGraph::deleteSupport(CubeID & supported) {
    auto support_it = cube_in_edges.find(supported);
    if(support_it == cube_in_edges.end()) {
      // We already don't have one
      return;
    } else {
      auto old_range = getInEdges(supported);
      SupportSet old_ss(old_range.first, old_range.second);
      for(CubeID from_id : old_ss) {
        removeEdge(from_id, supported);
      }
    }
    cube_in_edges.erase(support_it);
    assert(cube_in_edges.find(supported) == cube_in_edges.end());
  }

  std::pair<SupportSet::const_iterator, SupportSet::const_iterator>
  SupportGraph::getInEdges(const Cube & cube) const {
    assert(cube_db.isKnown(cube));
    CubeID id = cube_db.getID(cube);
    return getInEdges(id);
  }
  
  std::pair<SupportSet::const_iterator, SupportSet::const_iterator>
  SupportGraph::getInEdges(CubeID id) const {
    static SupportSet null_support;
    assert(cube_db.isKnown(id));
    auto it = cube_in_edges.find(id);
    if(it != cube_in_edges.end()) {
      return std::make_pair(it->second.begin(), it->second.end());
    } else {
      return std::make_pair(null_support.end(), null_support.end());
    }
  }

  std::pair<SupportSet::const_iterator, SupportSet::const_iterator>
  SupportGraph::getOutEdges(const Cube & cube) const {
    assert(cube_db.isKnown(cube));
    CubeID id = cube_db.getID(cube);
    return getOutEdges(id);
  }
  
  std::pair<SupportSet::const_iterator, SupportSet::const_iterator>
  SupportGraph::getOutEdges(CubeID id) const {
    static SupportSet null_support;
    assert(cube_db.isKnown(id));
    auto it = cube_out_edges.find(id);
    if(it != cube_out_edges.end()) {
      return std::make_pair(it->second.begin(), it->second.end());
    } else {
      return std::make_pair(null_support.end(), null_support.end());
    }
  }

  std::pair<std::set<CubeID>::const_iterator, std::set<CubeID>::const_iterator> 
  SupportGraph::getSupportedCubeIDs() const {
    return std::make_pair(supported_cubes.begin(), supported_cubes.end());
  }

  std::pair<std::set<CubeID>::const_iterator, std::set<CubeID>::const_iterator> 
  SupportGraph::getKnownCubeIDs() const {
    return std::make_pair(known_cubes.begin(), known_cubes.end());
  }

  bool SupportGraph::hasSupportSet(const Cube & cube) const {
    assert(cube_db.isKnown(cube));
    CubeID id = cube_db.getID(cube);
    return hasSupportSet(id);
  }

  bool SupportGraph::hasSupportSet(CubeID id) const {
    return cube_in_edges.find(id) != cube_in_edges.end();
  }
  
  std::set<Cube> SupportGraph::getSupportSet(const Cube & cube) const {
    assert(hasSupportSet(cube));
    CubeID id = cube_db.getID(cube);
    std::set<Cube> supports;
    auto it = cube_in_edges.find(id);
    for (SupportSet::iterator supp_it=it->second.begin(); supp_it!=it->second.end(); supp_it++){
      supports.insert(cube_db.getCube(*supp_it));
    }
    return supports;
  }

  size_t SupportGraph::getInDegree(const Cube & cube) const {
    assert(cube_db.isKnown(cube));
    CubeID id = cube_db.getID(cube);
    return getInDegree(id);
  }

  size_t SupportGraph::getInDegree(CubeID id) const {
    assert(cube_db.isKnown(id));
    auto it = cube_in_edges.find(id);
    if(it != cube_in_edges.end()) {
      return it->second.size();
    } else {
      return 0;
    }
  }

  bool SupportGraph::isBad(const Cube & cube) const {
    assert(cube_db.isKnown(cube));
    CubeID id = cube_db.getID(cube);
    return isBad(id);
  }

  bool SupportGraph::isBad(CubeID id) const {
    assert(cube_db.isKnown(id));
    return cube_db.getLemma(id).bad;
  }

  size_t SupportGraph::getOutDegree(const Cube & cube) const {
    assert(cube_db.isKnown(cube));
    CubeID id = cube_db.getID(cube);
    return getOutDegree(id);
  }

  size_t SupportGraph::getOutDegree(CubeID id) const {
    assert(cube_db.isKnown(id));
    auto it = cube_out_edges.find(id);
    if(it != cube_out_edges.end()) {
      return it->second.size();
    } else {
      return 0;
    }
  }

  int SupportGraph::getLevel(const Cube & cube) const {
    assert(cube_db.isKnown(cube));
    CubeID id = cube_db.getID(cube);
    return getLevel(id);
  }

  int SupportGraph::getLevel(CubeID id) const {
    assert(cube_db.isKnown(id));
    const Lemma & lem = cube_db.getLemma(id);
    return lem.lv;
  }

  /***
  Get the minimum level of all cubes in the given cube's support set.
  Returns 0 if the cube has no support set. 
  ***/
  int SupportGraph::getMinSupportLevel(const Cube & cube) const {
    CubeID id = cube_db.getID(cube);
    return getMinSupportLevel(id);
  }
  
  int SupportGraph::getMinSupportLevel(CubeID id) const {
    if (!hasSupportSet(id))
      return 0;
    
    auto it = cube_in_edges.find(id);
    int min_lv = INT_MAX;
    for (SupportSet::iterator supp_it=it->second.begin(); supp_it!=it->second.end(); supp_it++){
      min_lv = min(min_lv, getLevel(*supp_it));
    }
    return min_lv;
  }
  
  void SupportGraph::findUnsupportedCubes(const Cube & init_cube, std::set<CubeID> & unsupported_cubes) const {
    auto known_range = getKnownCubeIDs();

    // Not as efficient as possible but good enough
    bool found_new = true;
    while(found_new) {
      found_new = false;

      for(auto cube_it = known_range.first; cube_it != known_range.second; ++cube_it) {
        CubeID cube_id = *cube_it;

        if(unsupported_cubes.find(cube_id) != unsupported_cubes.end()) {
          continue;
        }

        if(getLevel(cube_id) == 0) {
          const Cube & cube = cube_db.getCube(cube_id); 
          assert(cube.size() == 1);
          ID lit = cube[0];

          if(std::find(init_cube.begin(), init_cube.end(), lit) == init_cube.end()) {
            unsupported_cubes.insert(cube_id);
            found_new = true;
          }
        } else if(hasSupportSet(cube_id)) {
          auto support_range = getInEdges(cube_id);

          for(auto support_it = support_range.first; support_it != support_range.second; ++support_it) {
            CubeID supporting_cube_id = *support_it;

            if(unsupported_cubes.find(supporting_cube_id) != unsupported_cubes.end()) {
              unsupported_cubes.insert(cube_id);
              found_new = true;
            }
          }
        } else {
          // We don't know a support set for this cube
          unsupported_cubes.insert(cube_id);
          found_new = true;
        }
      }
    }
  }

  void SupportGraph::buildInductiveTrace(const Cube & init_cube, vector<CubeSet> & frames) const {
    (void) init_cube;
    (void) frames;
    throw std::logic_error("buildInductiveTrace is no longer implemented");
  }

  void SupportGraph::writeDot(const std::string & file, const Cube & property) const {
    std::ofstream dot_file;
    dot_file.open(file, ios::out);
    writeDot(dot_file, property);
    dot_file.close();
  }

  void SupportGraph::writeDot(std::ostream & os, const Cube & property) const {
    CubeID property_id = CUBE_ID_NULL;
    if(!property.empty()) {
      assert(cube_db.isKnown(property));
      property_id = cube_db.getID(property);
    }

    SupportGraphDotWriter writer(*this, property_id);
    writer.writeDot(os);
  }

  void SupportGraph::writeInfDot(const std::string & file, const Cube & property) const {
    std::ofstream dot_file;
    dot_file.open(file, ios::out);
    writeInfDot(dot_file, property);
    dot_file.close();
  }

  void SupportGraph::writeInfDot(std::ostream & os, const Cube & property) const {
    CubeID property_id = CUBE_ID_NULL;
    if(!property.empty()) {
      assert(cube_db.isKnown(property));
      property_id = cube_db.getID(property);
    }

    SupportGraphDotWriter writer(*this, property_id);
    writer.writeInfDot(os);
  }

  void SupportGraph::computeStats(const Cube & property, bool inf_only){
    total=0;
    total_cnt=0;
    in_deg_dist.clear();
    out_deg_dist.clear();
    auto known_cube_range = getKnownCubeIDs();
    auto supp_cube_range = getSupportedCubeIDs();
    
    //count number of cubes in a level lower than infinity with no support set
    unknown_supp=0;
    for (auto it=known_cube_range.first; it!=known_cube_range.second; it++){
      if (getLevel(*it) >0 && !hasSupportSet(*it))
        unknown_supp++;
    }
    
    CubeID property_id = CUBE_ID_NULL;
    unordered_map<CubeID, unsigned> outdegrees; 
    set<CubeID> transitive_support;
    if(!property.empty() && cube_db.isKnown(property)) {
      property_id = cube_db.getID(property); 
      // TODO: this should not construct a dot writer
      SupportGraphDotWriter writer(*this, property_id);
      outdegrees = writer.getOutDegrees();
      transitive_support = writer.findTransitiveSupport(cube_db.getID(property));
    }

    
    //compute frequency distribution of indegrees
    for (auto it=supp_cube_range.first; it!=supp_cube_range.second; it++){
      if (inf_only && getLevel(*it) != INT_MAX) continue;
      if (getLevel(*it) == -1) continue;
      
      //count number of cubes and supports (in degree)
      int num_supports = cube_in_edges[*it].size();
      total += num_supports;
      total_cnt++;

      while ((int)in_deg_dist.size() <= num_supports)
        in_deg_dist.push_back(0);
      in_deg_dist[num_supports]++;
    }
    
    //compute frequency distribution of outdegrees
    for (auto it=outdegrees.begin(); it!=outdegrees.end(); it++){
      if (inf_only && getLevel(it->first) != INT_MAX) continue;
      if (getLevel(it->first) == -1) continue;
      while (out_deg_dist.size() <= it->second)
        out_deg_dist.push_back(0);
      out_deg_dist[it->second]++;
    }
    
    needed_inits = 0;
    for (auto it = transitive_support.begin(); it != transitive_support.end(); it++){
      const Lemma & lem = cube_db.getLemma(*it);
      if (lem.is_init) {
        needed_inits++;
      }
    }
  }
  
  void SupportGraph::printStats(bool inf_only) {
    cout<<"\n";
    cout<<(inf_only ? "Inf " : "")<<"Support Graph Statistics:"<<endl;
    if (!inf_only) {
      cout<<"Number of known cubes: "<<known_cubes.size()<<endl;
      cout<<"Number of supported cubes: "<<supported_cubes.size()<<endl;
      cout<<"Number of unsupported cubes (excluding initial): "<<unknown_supp<<endl;
    }
    cout<<"Average support set size: "<<(double)total/total_cnt<<endl;
    cout<<"Number of cubes: "<<total_cnt<<endl;
    cout<<"Frequency distribution for in degree (cubes with support set only):";

    // TODO: compute median support in computeStats
    int median_support = -1;
    int total_cubes = 0;
    int total_nonzero = total_cnt;
    if(in_deg_dist.size() > 0) {
     total_nonzero -= in_deg_dist[0];
    }
    for (int i = 0; i < (int) in_deg_dist.size(); i++) {
      cout << " " << in_deg_dist[i];
      if(i > 0) {
        total_cubes += in_deg_dist[i];
        if(median_support < 0 && total_cubes >= total_nonzero / 2) {
          median_support = i;
        }
      }
    }
    cout<<endl;
    
    cout<<"Frequency distribution for out degree (all known cubes):";
    for (int i=0; i<(int)out_deg_dist.size(); i++) {
      cout<<" "<<out_deg_dist[i];
    }
    cout << endl;
    cout << "Average number of support cubes: " << (double)total / total_nonzero << endl;
    cout << "Median number of support cubes: " << median_support << endl;
    if (!inf_only)
      cout<<"Number of initial state literals needed to support the property: "<<needed_inits<<endl;
  }
  
  bool CubeDatabase::isKnown(const Cube & cube) const {
    bool res = (cube_to_id.find(cube) != cube_to_id.end());
#ifndef NDEBUG
    if(res) {
      CubeID id = cube_to_id.at(cube);
      assert(id_to_cube.at(id) == cube);
    }
#endif
    return res;
  }

  bool CubeDatabase::isKnown(CubeID id) const {
    bool res = id_to_cube.find(id) != id_to_cube.end();
#ifndef NDEBUG
    if(res) {
      const Cube & cube = id_to_cube.at(id);
      assert(cube_to_id.at(cube) == id);
    }
#endif
    return res;
  }

  CubeID CubeDatabase::addLemma(const Lemma & lem) {
    const Cube cube = lem.getCube();
    if(isKnown(cube)) {
      return getID(cube);
    }

    CubeID id = getNextID();

    addCubeWithID(cube, id, lem);

    return id;
  }

  CubeID CubeDatabase::getID(const Cube & cube) const {
    assert(isKnown(cube));

    return cube_to_id.at(cube);
  }

  const Cube & CubeDatabase::getCube(CubeID id) const {
    assert(isKnown(id));

    return id_to_cube.at(id); 
  }

  Lemma & CubeDatabase::getLemma(const Cube & cube) {
    return const_cast<Lemma &>(static_cast<const CubeDatabase &>(*this).getLemma(cube));
  }

  Lemma & CubeDatabase::getLemma(CubeID id) {
    return const_cast<Lemma &>(static_cast<const CubeDatabase &>(*this).getLemma(id));
  }

  const Lemma & CubeDatabase::getLemma(const Cube & cube) const {
    assert(isKnown(cube));
    CubeID id = getID(cube);
    return getLemma(id);
  }

  const Lemma & CubeDatabase::getLemma(CubeID id) const {
    return lemma_db.at(id);
  }

  QuipSolver::QuipSolver(Model & m, QuipOptions o) : 
                                           m(m), 
                                           opts(o),
                                           logger(m.verbosity()),
                                           ev(NULL),
                                           sat_man(NULL),
                                           lift_man(NULL),
                                           sg_man(NULL), 
                                           sat_view(NULL), 
                                           lift(NULL),
                                           sg_view(NULL),
                                           cube_db(NULL),
                                           support_graph(NULL),
                                           current_level(0),
                                           next_sg_lit_index(0),
                                           sg_man_has_inf_act(false)
  {
    stats = QuipStats();

    cube_db = new CubeDatabase;
    support_graph = new SupportGraph(*cube_db, logger);
 
    prepare();
 }

  QuipSolver::~QuipSolver()
  {
    cleanup();

    assert(support_graph);
    delete support_graph;
    support_graph = NULL;
    
    assert(cube_db);
    delete cube_db;
    cube_db = NULL;
  }

  ID QuipSolver::getNPI() const { 
    const ExprAttachment * eat = (ExprAttachment *) model().constAttachment(Key::EXPR);

    ID npi = eat->outputFnOf(eat->outputs()[0]);
    Cube cube;
    Expr::conjuncts(*ev, npi, cube);
    npi = ev->apply(Expr::And, cube);

    model().constRelease(eat);

    return npi;
  }

  void QuipSolver::cleanupSAT() {
    if(sat_view != NULL) {
      delete sat_view;
      sat_view = NULL;
    }

    if(lift != NULL) {
      delete lift;
      lift = NULL;
    }

    if(sg_view != NULL) {
      delete sg_view;
      sg_view = NULL;
    }

    if(sat_man != NULL) {
      delete sat_man;
      sat_man = NULL;
    }

    if(lift_man != NULL) {
      delete lift_man;
      lift_man = NULL;
    }

    if(sg_man != NULL) {
      delete sg_man;
      sg_man = NULL;
    }
  }

  void QuipSolver::cleanupTR() {
    tr_clauses.clear();
  }

  void QuipSolver::cleanup() {
    if(ev) {
      delete ev;
      ev = NULL;
    } 

    cleanupSAT();
    cleanupCTPs();
    cleanupTR();
  }

  /*
  Rebuild the inductive trace from the support graph, considering the
  current initial state 
  */
  void QuipSolver::rebuildTrace() { 
    logger.verbose() << "Rebuilding inductive trace from support graph" << std::endl;
    throw std::logic_error("rebuildTrace is no longer implemented");
  }

  /***
  Grow the inductive trace to size k if needed
  ***/
  void QuipSolver::growFrames(size_t k) {
    assert(frames.size() == active_ctps.size());
    if(frames.size() < k) {
      frames.resize(k);
      active_ctps.resize(k);
    }
  } 

  void QuipSolver::cleanupCTPs() {
    for (size_t k=0; k < active_ctps.size(); k++){
      for (list<CTP*>::iterator it=active_ctps[k].begin(); it!=active_ctps[k].end(); it++){
        delete (*it);
      }
    }
    active_ctps.clear();
  }

  void QuipSolver::prepare(QuipOptions * new_opts, bool clear_finf, bool rebuild_trace) {
    if(new_opts) {
      opts = *new_opts;
    }

    cleanup();

    assert(!ev);
    ev = m.newView();

    verbosity = m.verbosity();
    logger.setVerbosity(m.verbosity());

    const ExprAttachment * eat = (ExprAttachment *) model().constAttachment(Key::EXPR);
    state_vars = eat->stateVars();
    input_vars = eat->inputs();

    init_cube = eat->initialConditions();
    initially.clear();
    if(opts.init_cube) {
      initially.insert(opts.init_cube->begin(), opts.init_cube->end());
    } else {
      initially.insert(init_cube.begin(), init_cube.end());
    }
    stats.init_states = initially.size();
 
    m.constRelease(eat);

    // Handle the absence of a COIAttachment gracefully by assuming all
    // registers are in the COI 
    coi_latches.clear();
    const COIAttachment * cat = (COIAttachment const *) model().constAttachment(Key::COI);
    if(cat) {
      COI coi = cat->coi();

      COI::range latch_range = coi.cCOI();
      coi_latches.insert(latch_range.first, latch_range.second);

      model().constRelease(cat);
    } else {
      coi_latches.insert(state_vars.begin(), state_vars.end());
    }

    rsman.clear();
    
    if(clear_finf) {  
      for(auto it = inf_cubes.begin(); it != inf_cubes.end(); ++it) {
        support_graph->deleteCube(*it);  
      }
      inf_cubes.clear();
    }

    if(opts.sg_type != NO_SG && rebuild_trace) {
      rebuildTrace();
    } else {
      frames.clear();
    } 

    initTR();

    // Set up F_0
    initFrames();
  }

  std::string QuipSolver::scube(CubeID id) const {
    return stringOfClauseOrCube(cube_db->getCube(id));
  }

  std::string QuipSolver::stringOfClauseOrCube(const vector<ID> & clause_or_cube, const std::string & sep) const {
    assert(ev);
    std::stringstream ss;
    for (auto it = clause_or_cube.begin(); it != clause_or_cube.end(); ++it){
      ss << " " << Expr::stringOf(*ev, *it);
      if((it+1) != clause_or_cube.end() && sep != "") {
        ss << " " << sep;
      }
    }
    std::string result = ss.str();
    boost::trim(result);
    return result;
  }

  /***
  Return true if a subsumes b, i.e. b is a subset of a.
  Cubes are assumed to be sorted.
  ***/
  bool QuipSolver::subsumes(const Cube & a, const Cube & b) 
  {
    size_t i=0, j=0;
    for (; j<b.size(); j++){
      while (i<a.size() && a[i] != b[j]){
        i++;
      }
      if (i == a.size()) return false;
    }
    return true;
  }

  /***
  Return the ID of the level variable. In keeping with IC3,
  levelVar(-1) is the property lemma.
  ***/
  ID QuipSolver::levelVar(int lv) const 
  {  
    ostringstream buf;
    if(lv == -1) {
      buf << "bad";
    } else {
      buf << "lvl" << lv;
    }
    return ev->newVar(buf.str());
  }
  
  ID QuipSolver::var(ID lit) const {
    Expr::Op op = ev->op(lit);
    if (op == Expr::Var)
      return lit;
    else {
      assert (op == Expr::Not);
      return var(ev->apply(Expr::Not, lit));
    }
  }

  ID QuipSolver::newSGVar(const Cube & cube) {
    assert(isKnownLemma(cube));
    CubeID id = cube_db->getID(cube);
    return newSGVar(id);
  }

  ID QuipSolver::newSGVar(CubeID id) {
    assert(cube_to_sg_lit.find(id) == cube_to_sg_lit.end());
    std::ostringstream buf;
    buf <<  "sg_act_" << next_sg_lit_index;
    next_sg_lit_index++;
    ID lit = ev->newVar(buf.str());
    cube_to_sg_lit[id] = lit;
    assert(sg_lit_to_cube.find(lit) == sg_lit_to_cube.end());
    sg_lit_to_cube[lit] = id;
    return lit;
  }

  ID QuipSolver::SGVar(const Cube & cube) {
    assert(isKnownLemma(cube));
    CubeID id = cube_db->getID(cube);
    auto map_it = cube_to_sg_lit.find(id);
    if(map_it == cube_to_sg_lit.end()) {
      return newSGVar(id);
    } else {
      return map_it->second;
    }
  }

  /***
  Returns true if the given literal is a support graph activation literal.
  It must have positive polarity for this to return true.
  ***/
  bool QuipSolver::isSGVar(ID lit) const {
    auto it = sg_lit_to_cube.find(lit);
    return it != sg_lit_to_cube.end();
  }

  const Cube & QuipSolver::cubeFromSGVar(ID lit) const {
    assert(sg_lit_to_cube.find(lit) != sg_lit_to_cube.end());
    CubeID id = sg_lit_to_cube.at(lit);
    return cube_db->getCube(id);
  }

  Lemma & QuipSolver::getLemma(const Cube & cube) {
    // const_cast is the prefered way to avoid duplication of code between
    // const and non-const versions in this case
    return const_cast<Lemma &>(static_cast<const QuipSolver &>(*this).getLemma(cube));
  }

  Lemma & QuipSolver::getLemma(CubeID id) {
    return const_cast<Lemma &>(static_cast<const QuipSolver &>(*this).getLemma(id));
  }

  const Lemma & QuipSolver::getLemma(const Cube & cube) const {
    assert(cube_db->isKnown(cube));
    return cube_db->getLemma(cube);
  }

  const Lemma & QuipSolver::getLemma(CubeID id) const {
    return cube_db->getLemma(id);
  }

  bool QuipSolver::isKnownLemma(const Cube & cube) const {
    return cube_db->isKnown(cube);
  }

  void QuipSolver::markBad(const Cube & cube) {
    assert(isKnownLemma(cube));
    Lemma & lem = getLemma(cube);
    lem.bad = true;
  }

  void QuipSolver::markUgly(const Cube & cube) {
    assert(isKnownLemma(cube));
    Lemma & lem = getLemma(cube);
    if (!lem.ugly) { stats.ugly_marked++; }
    lem.ugly = true;
  }

  void QuipSolver::unmarkUgly(const Cube & cube) {
    assert(isKnownLemma(cube));
    Lemma & lem = getLemma(cube);
    if (lem.ugly) { stats.ugly_unmarked++; }
    lem.ugly = false;
  }

  /***
  Checks if the lemma is acceptable. For now, acceptable means not bad
  and not ugly, but we could easily consider other critera
  ***/
  bool QuipSolver::isAcceptable(const Cube & cube) const {
    assert(isKnownLemma(cube));
    CubeID id = cube_db->getID(cube);
    return isAcceptable(id);
  }

  bool QuipSolver::isAcceptable(CubeID id) const {
    const Lemma & lem = getLemma(id);
    return !lem.bad && (!lem.ugly || !opts.ugly_lemmas);
  }

  /***
  Returns true if and only if we know a support set for cube that 
  is acceptable. If cube is not a known lemma, returns false
  ***/
  bool QuipSolver::hasAcceptableSupport(const Cube & cube) const {
    if (!isKnownLemma(cube)) {
      return false;
    }

    bool has_support = support_graph->hasSupportSet(cube);
    if (!has_support) {
      return false;
    }

    auto support_range = support_graph->getInEdges(cube);
    for (auto it = support_range.first; it != support_range.second; ++it) {
      CubeID id = *it;
      if (!isAcceptable(id)) {
        return false;
      }
    }

    return true;
  }

  bool QuipSolver::hasBadSupport(const Cube & cube) const {
    if (!isKnownLemma(cube)) {
      return false;
    }

    bool has_support = support_graph->hasSupportSet(cube);
    if (!has_support) {
      return false;
    }

    auto support_range = support_graph->getInEdges(cube);
    for (auto it = support_range.first; it != support_range.second; ++it) {
      CubeID id = *it;
      if (getLemma(id).bad) {
        return true;
      }
    }

    return false;
  }

  /***
  Checks if cube blocks any CTPs at level k and set those CTP's blocked value to true.
  This function should be called whenever a new lemma is added to frame k.
  min_check: minimum level of CTPs to check. Useful if it's already known that the cube 
  doesn't block any CTPs at some lower level (eg lv - 1)
  ***/
  void QuipSolver::checkCTPs(const Cube & cube, int lv, int min_check)
  {
    //Need to check all levels <= lv, since we're actually adding the cube to all of them
    for (int k=min_check; k<=min(lv,(int)active_ctps.size()-1); k++){
      for (list<CTP*>::iterator it=active_ctps[k].begin(); it!=active_ctps[k].end(); ){
        if ((*it)->can_delete){
          //this means the CTP has been read as blocked by the lemma it counterexamples
          assert((*it)->blocked);
          delete (*it);
          it = active_ctps[k].erase(it);
        }
        else {
          if ((*it) && !(*it)->blocked){
            stats.ctp_subsummption_checks++;
            if (subsumes((*it)->state, cube)){
              logger.logorrheic() << "Marking CTP "<< pc((*it)->state) << " as blocked" << endl;
              (*it)->blocked = true;
            }
          }
          it++;
        }
      }
    }
  }
  
  /***
  Add cube to sat manager at level lv.
  ***/
  void QuipSolver::addCubeToManager(const Cube & cube, int lv, SAT::Manager * manager)
  {
    //convert to clause
    Clause cls;
    Expr::complement(*ev, cube, cls);
    //add the level variable
    if (lv != INT_MAX){
      cls.push_back(ev->apply(Expr::Not, levelVar(lv)));
    }
    if(manager) {
      manager->add(cls);
    } else {
      sat_man->add(cls);
    }
  }

  /***
  Add cube to SAT manager with its unique per-cube activation literal
  (or without it, if act = false)
  ***/
  void QuipSolver::addCubeToSGManager(const Cube & cube, bool act, SAT::Manager * manager) {
    // Add it to the support graph manager
    SAT::Manager * dst = manager ? manager : sg_man;

    Cube cube_copy(cube);
    if(act) {
      ID nsg_lit = ev->apply(Expr::Not, SGVar(cube_copy));
      cube_copy.push_back(nsg_lit);
    }

    addCubeToManager(cube_copy, INT_MAX, dst);
  }

  
  void QuipSolver::addCubeSetToManager(const CubeSet & cubes, int lv, SAT::Manager * manager)
  {
    for (auto it = cubes.begin(); it != cubes.end(); it++){
      const Cube & cube = *it;
      assert(isKnownLemma(cube));
      addCubeToManager(cube, lv, manager);
    }
  }

  void QuipSolver::addCubeSetToSGManager(const CubeSet & cubes, bool act, SAT::Manager * manager)
  {
    for (auto it = cubes.begin(); it != cubes.end(); it++){
      const Cube & cube = *it;
      assert(isKnownLemma(cube));
      addCubeToSGManager(cube, act, manager);
    }
  }

  void QuipSolver::logAddCube(const std::string & prefix, 
                              int lv, 
                              const Cube & cube, 
                              std::ostream & stream) const 
  {
    CubeID id = cube_db->getID(cube);
    std::string lv_string = (lv == INT_MAX ? "inf" : std::to_string(lv));
    stream << prefix << " " << lv_string << ": " 
           << pc(cube) << " [" << id << "]"
           << std::endl;
  }

  /***
  Adds the cubes to the inductive trace at lv and records this
  fact in the support graph. The cube should be sorted.
  Returns the previous level (if any, -1 otherwise)
  ***/ 
  int QuipSolver::addCubeToTrace(const Cube & cube, int lv) 
  {
    int prev_level = -1;
    if (lv < INT_MAX) {
      growFrames(lv+1);
    }

    // find frame which currently contains this lemma, if any
    assert(lv == INT_MAX || (size_t) lv < frames.size());
    for (int cur = min((int)frames.size() - 1, lv); cur >= 0; cur--){
      if (frames[cur].find(cube) != frames[cur].end()) {
        frames[cur].erase(cube);
        prev_level = cur;
        assert(isKnownLemma(cube));
        break;
      }
    }

    if (lv == INT_MAX) {
      inf_cubes.insert(cube);
    } else {
      assert(lv >= 0);
      assert(lv < (int)frames.size());

      frames[lv].insert(cube);
    }

    if (isKnownLemma(cube)) {
      logAddCube("To", lv, cube, logger.verbose());
      Lemma & lem = getLemma(cube);
      assert(lem.lv <= lv);
      // TODO: if (lem.bad), we must've just learned a bad lemma (since
      // we never push bad lemmas) - seems like a good stat to record
      lem.lv = lv;
    } else {
      Lemma lem = Lemma(cube, lv);
      if(lv == 0 && cube != getCurrentBad()) {
        lem.is_init = true;
      }
      cube_db->addLemma(lem);
      logAddCube("At", lv, cube, logger.verbose());
    }

    return prev_level;
  }

  /***
  Add lemma for the given cube to frames and add to sat manager.
  ***/
  void QuipSolver::addCube(const Cube & cube, int lv)
  {
    assert(!cube.empty());
    Cube cube_sorted(cube);
    sort(cube_sorted.begin(), cube_sorted.end());

    int prev_lv = addCubeToTrace(cube_sorted, lv);
    unmarkUgly(cube_sorted);

    // add to sat manager for all levels < lv
    for (int k = std::max(prev_lv + 1, 0); k < lv && lv < INT_MAX; k++) {
      addCubeToManager(cube, k);
    }
    addCubeToManager(cube, lv);

    // Add it to the support graph manager
    // No activation literal if it's at infinity
    bool with_act = lv < INT_MAX;
    addCubeToSGManager(cube, with_act);
    
    if(lv == INT_MAX) {
      //lift_man contains all level inf cubes
      addCubeToManager(cube_sorted, lv, lift_man);
    }

    if (opts.use_ctps){
      checkCTPs(cube_sorted, lv);
    }
  }
  
  /***
  Build a set of assumptions that enables any cubes in F_{lvl} and disable
  all others in a Support Graph query.
  If acceptable_only is true, unacceptable lemmas (such as bad ones) are 
  excluded from F_{lvl}
  Infinity lemmas are added without activation literals, so they are not
  considered here and are always included in F_*
  ***/
  Cube QuipSolver::getSGAssumps(int lvl, bool acceptable_only) {
    Cube assumps;
    // TODO remove or #ifdef guard the debug checks
    std::set<ID> assumps_set;
    for(size_t i = 0; i < frames.size(); ++i) {
      const CubeSet & cubes = frames.at(i);
      for(auto it = cubes.begin(); it != cubes.end(); ++it) {
        const Cube & cube = *it;
        assert(isKnownLemma(cube));

        ID sg_lit = SGVar(cube);
        ID nsg_lit = ev->apply(Expr::Not, sg_lit);

        if ((int)i >= lvl) {
          // The lemmas are in the frame, set the sg_lit to 0
          // if it's acceptable or we're allowing unacceptables
          if(!acceptable_only || isAcceptable(cube)) {
            assumps.push_back(nsg_lit);
          }
          // Still add it to the set for error-checking purposes
          assumps_set.insert(nsg_lit);

          assert(assumps_set.find(sg_lit) == assumps_set.end());
        } else {
          // The lemmas are not in the frame, set the sg_lit to 1
          // We don't need to set it to one and seemingly gain a lot
          // performance by not doing so
          //assumps.push_back(sg_lit);
          assumps_set.insert(sg_lit);

          assert(assumps_set.find(nsg_lit) == assumps_set.end());
        }
      }
    }

    if(sg_man_has_inf_act) {
      for(auto it = inf_cubes.begin(); it != inf_cubes.end(); ++it) {
        const Cube & cube = *it;
        assert(isKnownLemma(cube));

        ID sg_lit = SGVar(cube);
        ID nsg_lit = ev->apply(Expr::Not, sg_lit);

        assumps.push_back(nsg_lit);
        assumps_set.insert(nsg_lit);

        assert(assumps_set.find(sg_lit) == assumps_set.end());
      }
    }

    return assumps;
  }

  void QuipSolver::computeSupportSetFromCrits(const Cube & crits, std::set<Cube> & support) const {
    for(auto it = crits.begin(); it != crits.end(); ++it) {
      ID crit = *it;
      ID ncrit = ev->apply(Expr::Not, crit);

      if(isSGVar(ncrit)) {
        // In support set
        const Cube & support_cube = cubeFromSGVar(ncrit);
        support.insert(support_cube);
      }
    } 
  }

  int QuipSolver::getSupportLevel(const Cube & cube) const {
    assert(isKnownLemma(cube));
    int lvl = getLemma(cube).lv;

    if(lvl < INT_MAX) {
      lvl -= 1;
    }

    return lvl;
  }

  /***
  Tries to compute an acceptable (no blacklisted lemmas) support set.
  Returns true if an acceptable support set was found, false otherwise
  ***/
  bool QuipSolver::computeAcceptableSupport(const Cube & cube, int lvl) {
    stats.support_calls++;
    QuipAutoTimer auto_timer(stats.support_time);
    if(lvl < 0) {
      lvl = getSupportLevel(cube);
    }
    assert(lvl >= 0);
    bool cons = consecutionActivation(cube, lvl, NULL, NULL, CONSECUTION_ACCEPTABLE_SG);
    return cons;
  }

  void QuipSolver::computeSupport(const Cube & cube) {
    int lvl = getSupportLevel(cube);
    assert(lvl >= 0);
    computeSupport(cube, lvl);
  }

  /***
  Computes the support set of the cube at the given level and stores it in
  the support graph. Just a simple wrapper around consecution.
  ***/
  void QuipSolver::computeSupport(const Cube & cube, int lvl) {
    bool cons = consecution(cube, lvl, NULL, NULL, true); 
    assert(cons);
  }

  /***
  This function is called after a satisified sat query. It gets the state, inputs, 
  and primed inputs from the satisfying assignment which are used to construct the 
  counterexample trace.
  ***/
  void QuipSolver::getCubesFromAsgn(const SAT::Assignment & asgn, 
                                          Cube * cube, 
                                          Cube * input_cube, 
                                          Cube * p_input_cube,
                                          Cube * p_state) const
  {
    ExprAttachment const * const eat = (ExprAttachment const *) model().constAttachment(Key::EXPR);
    for (SAT::Assignment::const_iterator it = asgn.begin(); it != asgn.end(); ++it){
      if (it->second != SAT::Unknown){
        assert(ev->nprimes(it->first) <= 1); 
        if (ev->nprimes(it->first) == 0 && eat->isInput(it->first)){
          //regular input
          if (input_cube != NULL){
            input_cube->push_back(it->second == SAT::False ?
              ev->apply(Expr::Not, it->first) :
              it->first);
          }
        }
        else if (ev->nprimes(it->first) == 1 && eat->isInput(ev->unprime(it->first))){
          //primed input
          if (p_input_cube != NULL){
            p_input_cube->push_back(it->second == SAT::False ?
              ev->apply(Expr::Not, it->first) :
              it->first);
          }
        }
        else if(ev->nprimes(it->first) == 1 && eat->isStateVar(ev->unprime(it->first))) {
          // primed sate
          if(p_state != NULL) {
            ID lit = it->first;
            ID unprimed_lit = ev->unprime(lit);
            p_state->push_back(it->second == SAT::False ?
              ev->apply(Expr::Not, unprimed_lit) :
              unprimed_lit);
          }
        }
        else {
          //state variable
          assert(eat->isStateVar(it->first));
          assert(ev->nprimes(it->first) == 0);
          if(cube) {
            ID id = it->first;
            if(coi_latches.find(id) != coi_latches.end()) {
              ID nid = ev->apply(Expr::Not, id);
              cube->push_back(it->second == SAT::False ? nid : id);
            }
          }
        }
      }
    }
    model().constRelease(eat);
  }
  
  /***
  Check the case where bad_cube intersects the initial conditions.
  ***/
  bool QuipSolver::checkInitial(const Cube & bad_cube) const
  {  
    //do sat query F_0 ^ bad
    Expr::IDVector assumps;
    for(auto it = bad_cube.begin(); it != bad_cube.end(); ++it) {
      assumps.push_back(*it);
    }
    assumps.push_back(levelVar(0));
    
    stats.sat_calls++;
    bool is_sat = sat_view->sat(&assumps);
    
    if (is_sat){
      logger.informative() <<"Bad state intersects initial conditions"<<endl;
    }
    return is_sat;
  }

  /***
  Extract the support dependencies from the given unsat core and add them to 
  the support graph. Core should be given directly from the sat solver.
  Specifically, for each clause x in core, check if x is in frames[k] and if 
  so add the support x -> c. 
  ***/
  void QuipSolver::addCoreToGraph(const vector<Clause> & core, const Cube & c, int k)
  { 
    set<Cube> support_set; //all cubes supporting c
    logger.logorrheic()<<"Extracting support set from unsat core for cube " << pc(c) << endl;
    ID lvlk = levelVar(k);
    
    for (size_t i=0; i<core.size(); i++){
      //Figure out where core[i] came from - frame k, infcubes, or other
      Cube core_cube;
      Expr::complement(*ev, core[i], core_cube); //core from sat solver is clauses
      //cube will contain "lvlk" literal iff it came from Fk.
      //Do linear search even though core is sorted since we have to delete the 
      //lvlk literal anyway.
      bool found = false;
      //iterate in reverse since lvlk usually seems to be the last literal
      for (Cube::reverse_iterator rit=core_cube.rbegin(); rit!=core_cube.rend(); ++rit){
        if (*rit == lvlk){
          found = true;
          core_cube.erase(--(rit.base())); 
          assert(cube_db->isKnown(core_cube));
          support_set.insert(core_cube);
          logger.logorrheic() << "Cube (" << pc(core_cube) 
                                << ") supports cube ("
                                << pc(c)
                                << ")" << std::endl;
          break;
        }
      }
      
      //core_cube is not from F_k. Check if it's in inf_cubes. 
      if (!found && inf_cubes.find(core_cube) != inf_cubes.end()) {
        assert(cube_db->isKnown(core_cube));
        support_set.insert(core_cube);
        logger.logorrheic() << "Cube (" << pc(core_cube) 
                              << ") supports cube ("
                              << pc(c)
                              << ")" << std::endl;
      }
    }
    
    //build support graph using this core
    assert(cube_db->isKnown(c));
    support_graph->recordSupport(support_set, c);
    
    if (opts.print_stats){
      stats.core_cnt++;
      stats.core_size_total += (int)support_set.size();
      //calculate total number of active cubes at level k (includes inf)
      int total_cubes = 0;
      for (int i=k; i<(int)frames.size(); i++)
        total_cubes += (int)frames[i].size();
      total_cubes += (int)inf_cubes.size();
      stats.core_vs_frame_size += (double)support_set.size()/total_cubes;
    }
  }

  /***
  Initialzes the SAT::Assignment with latches in the COI and inputs
  ***/
  void QuipSolver::initializeAsgn(SAT::Assignment & asgn, bool primed_state) const {
    asgn.clear();

    for (auto it = coi_latches.begin(); it != coi_latches.end(); ++it) {
      asgn.insert(SAT::Assignment::value_type(*it, SAT::Unknown));
      if(primed_state) {
        asgn.insert(SAT::Assignment::value_type(ev->prime(*it), SAT::Unknown));
      }
    }
    for (size_t i=0; i<input_vars.size(); i++) {
      asgn.insert(SAT::Assignment::value_type(input_vars[i], SAT::Unknown));
      asgn.insert(SAT::Assignment::value_type(ev->prime(input_vars[i]), SAT::Unknown));
    }
  }

  /***
  Compute a reduced cube from the critical assumptions
  ***/
  void QuipSolver::computeReduced(const Cube & c, Cube & reduced, const Cube & crits) const {
    reduced.clear();

    Cube crits_sorted = crits;
    std::sort(crits_sorted.begin(), crits_sorted.end());

    for (auto it = c.begin(); it != c.end(); it++) {
      if (binary_search(crits_sorted.begin(), crits_sorted.end(),
                        Expr::primeFormula(*ev, *it))) {
        reduced.push_back(*it);
      }
    }

    if (!initiation(reduced)) {
      // Put back a literal in order to make reduced satisfy initiation
      for (auto it = c.rbegin(); it != c.rend(); it++) {
        ID lit = *it;
        ID nlit = ev->apply(Expr::Not, lit);
        if (initially.count(nlit)) {
          reduced.push_back(lit);
          std::sort(reduced.begin(), reduced.end());
          assert(initiation(reduced));
          break;
        }
      }
    }
    
    stats.cube_reduction_total += (double)reduced.size() / c.size();
    stats.cube_reduction_cnt++;
  }

  /*** 
  Do consecution using the SAT manager with activation literals for each 
  individual clause. Update the support graph if consecution holds
  ***/
  bool QuipSolver::consecutionActivation(const Cube & c, 
                                         int k, 
                                         SAT::Assignment * asgn, 
                                         Cube * reduced,
                                         ConsecutionType type) {
    bool do_support_graph = (type != CONSECUTION_NO_SG);
    assert(!do_support_graph || isLitBasedSG(opts.sg_type));
    bool acceptable_sg = (type == CONSECUTION_ACCEPTABLE_SG);

    // If the support graph demands SG on every consecution call, this 
    // may not be a "real" lemma. Add it to the database at level -1,
    // but not to the inductive trace
    if(do_support_graph && !isKnownLemma(c)) {
      assert(requiresSGOnConsecution(opts.sg_type));
      cube_db->addLemma(Lemma(c, -1));
    }

    Clause cc;
    Expr::complement(*ev, c, cc);
    SAT::GID gid = sg_view->newGID();
    sg_view->add(cc, gid);

    Expr::IDVector assumps, cassumps, cube_assumps, level_assumps;
    cube_assumps = c;
    // cube_assumps = c'
    primeIDVector(ev, cube_assumps);

    level_assumps = getSGAssumps(k, acceptable_sg);

    // Add the appropriate assumptions for the level
    assumps.insert(assumps.end(), level_assumps.begin(), level_assumps.end());
    cassumps = assumps;

    // only include the cube_assumps in the criticals if reducing
    if(reduced != NULL) {
      cassumps.insert(cassumps.end(), cube_assumps.begin(), cube_assumps.end());
    }
    assumps.insert(assumps.end(), cube_assumps.begin(), cube_assumps.end());

    stats.sat_calls++;
    bool is_sat = sg_view->sat(&assumps, asgn, &cassumps);
    sg_view->remove(gid);

    if(!is_sat && reduced) {
      computeReduced(c, *reduced, cassumps);
    }

    if(!is_sat && do_support_graph) {
      if(hasBadSupport(c)) {
        support_graph->deleteSupport(c);
      }
      std::set<Cube> support;
      computeSupportSetFromCrits(cassumps, support);
      for(auto it = support.begin(); it != support.end(); ++it) {
        const Cube & support_cube = *it;
        logger.logorrheic() << "Cube (" << pc(support_cube) 
                            << ") supports (" << pc(c) 
                            << ")" << std::endl;
        assert(!acceptable_sg || isAcceptable(support_cube));
      }
      support_graph->recordSupport(support, c);
    }

    return !is_sat;
  }
   
  /***
  Do consecution using the normal SAT manager with level variables.
  ***/
  bool QuipSolver::consecutionStandard(const Cube & c, 
                                       int k, 
                                       SAT::Assignment * asgn, 
                                       Cube * reduced, 
                                       ConsecutionType type) {

    bool do_support_graph = (type != CONSECUTION_NO_SG);

    // The UNSAT core method doesn't support acceptable SGs
    assert(type != CONSECUTION_ACCEPTABLE_SG);
    assert(!do_support_graph || isCoreBasedSG(opts.sg_type));

    //add c to sat view as group so it can be removed after
    Clause cc;
    Expr::complement(*ev, c, cc);
    SAT::GID gid = sat_view->newGID();
    sat_view->add(cc,gid);
    
    //add !c' as assumption
    Expr::IDVector assumps;
    Expr::IDVector cassumps;
    for (size_t i=0; i<c.size(); i++){
      assumps.push_back(ev->prime(c[i]));
    }
    cassumps = assumps;

    assumps.push_back(levelVar(k));

    bool should_do_core = do_support_graph && isCoreBasedSG(opts.sg_type);
    
    vector<Clause> core;
    stats.sat_calls++;
    bool is_sat = sat_view->sat(&assumps, asgn, 
                                reduced ? &cassumps : NULL, 
                                SAT::NULL_GID, 
                                false, 
                                NULL, 
                                should_do_core ? &core : NULL);
    sat_view->remove(gid);
    
    if(!is_sat && reduced) {
      computeReduced(c, *reduced, cassumps);
    }
    
    if (!is_sat && should_do_core) {   
      assert(sat_view->supportsCore());
      addCoreToGraph(core, c, k);
    }
    
    return !is_sat;
  }

  /***
  Do sat query F_k ^ c ^ T ^ !c'
  Return true if unsat, meaning c is inductive relative to F_k. 
  ***/
  bool QuipSolver::consecution(const Cube & c, int k, SAT::Assignment * asgn, Cube * reduced, bool do_support_graph)
  {    
    stats.consecution_calls++;

    bool do_sg = do_support_graph || requiresSGOnConsecution(opts.sg_type); 
    ConsecutionType type = do_sg ? CONSECUTION_SG : CONSECUTION_NO_SG;

    assert(!(do_sg && opts.sg_type == NO_SG && !opts.do_final_sg));

    bool use_act = do_sg && isLitBasedSG(opts.sg_type);
    
    if(do_sg && use_act) {
      return consecutionActivation(c, k, asgn, reduced, type);
    } else {
      return consecutionStandard(c, k, asgn, reduced, type);
    }
  }
  
  /***
  Syntactically check if cube intersects with initial state
  ***/
  bool QuipSolver::initiation(const Cube & cube) const {
    for (auto it = cube.begin(); it != cube.end(); ++it) {
      if (initially.find(ev->apply(Expr::Not, *it)) != initially.end()) {
        return true;
      }
    }
    return false;
  }
  
  /***
  The join procedure used in generalization
  ***/
  bool QuipSolver::join(Cube & cube, const SAT::Assignment & asgn, const set<ID> & keep) const 
  {
    vector<ID> rcube;
    for (vector<ID>::iterator it = cube.begin(); it != cube.end(); it++) {
      SAT::Assignment::const_iterator ait = asgn.find(var(*it));
      Expr::Op op = ev->op(*it);
      if ((op == Expr::Not && ait->second == SAT::True) ||
        (op == Expr::Var && ait->second == SAT::False)){
        if (keep.find(*it) != keep.end()){
          // about to drop a literal that should be kept
          return false;
        }
      }
      else {
        rcube.push_back(*it);
      }        
    }
    cube = rcube;
    return true;
  }
  
  /***
  The down procedure used in generalization
  ***/
  bool QuipSolver::down(Cube & cube, int k, const set<ID> & keep) {
    SAT::Assignment asgn;
    for (vector<ID>::iterator it = cube.begin(); it != cube.end(); it++)
      asgn.insert(SAT::Assignment::value_type(var(*it), SAT::Unknown));
    
    while (true) {
      //cout<<"down, k = "<<k<<", cube = "; printClauseOrCube(cube);
      if (!initiation(cube))
        return false;
      if (!consecution(cube, k, &asgn)) {
        /*cout<<"satisfying assignment:"<<endl;
        for (SAT::Assignment::iterator it=asgn.begin(); it!=asgn.end(); it++){
          cout<<Expr::stringOf(*ev, it->first)<<" = "<<it->second<<endl;
        }*/
        if (!join(cube, asgn, keep))
          return false;
      }
      else{
        return true;
      }
    }
  }
  
  /***
  Simplified version of mic from IC3
  ***/
  void QuipSolver::mic(Cube & cube, int k)
  {
    set<ID> keep;
    for (;;) {
      // try brute force reduction four times
      unsigned int i = 0;
      int cnt = 3;
      for (; i < cube.size() && cnt >= 0; ++i) {
        if(keep.find(cube[i]) != keep.end())
          continue;
        vector<ID> dcube(cube.size()-1);
        for (unsigned int j = 0, l = 0; j < cube.size(); ++j){
          if (j != i)
            dcube[l++] = cube[j];
        }
        if (down(dcube, k, keep)) {
          cube = dcube;
          break;
        }
        keep.insert(cube[i]);
        --cnt;
      }

      if (i == cube.size() || cnt < 0) return;
    }
  }
  
  /***
  Generalize the given cube by removing literals while maintaining inductiveness 
  (based on std_generalize from IC3).
  lo - lowest level to consider
  hi - highest level to consider
  ***/
  int QuipSolver::generalize(int lo, int hi, Cube & cube)
  {
    logger.logorrheic() <<"generalizing cube " << pc(cube) << std::endl;

    //Check if inductive relative to truly inductive clauses
    Cube reduced;
    if(consecution(cube, INT_MAX, NULL, &reduced)) {
      cube = reduced;
      mic(cube, INT_MAX);
      return INT_MAX;
    }

    // Binary search for highlest level relative to which the cube
    // is inductive. Once the level is found, reduce the cube using
    // its core, and continue on to call down, mic, etc.
    int k = lo;
    
    int _lo = lo, _hi = hi;
    Cube successful_cube;
    while(_lo <= _hi) {
      int check_k = (_lo + _hi) / 2;
      if(consecution(cube, check_k, NULL, &reduced)) {
        k = check_k;
        _lo = check_k + 1;
        successful_cube = reduced;
        if(opts.under_generalize) {
          cube = reduced;
        }
        reduced.clear();
      } else {
        _hi = check_k - 1;
      }
    }
    
    if(k > lo) {
      assert(!successful_cube.empty());
      cube = successful_cube;
    }
    
    // use down to push forward as far as possible
    set<ID> dummy;
    for (; k < hi; ++k) {
      vector<ID> ccube(cube);
      if (down(ccube, k+1, dummy))
        cube = ccube;
      else
        break;
    }

    if (k == hi) {
      //Check if inductive relative to truly inductive clauses
      vector<ID> ccube(cube);
      if(down(ccube, INT_MAX, dummy)) {
        cube = ccube;
        mic(cube, INT_MAX);
        return INT_MAX;
      }
    }

    //Even if it wasn't inductive relative to truly inductive clauses, the
    //generalized clause might be
    if(consecution(cube, INT_MAX, NULL, &reduced)) {
      cube = reduced;
      mic(cube, INT_MAX);
      return INT_MAX;
    }

    // definitely inductive
    mic(cube, k);
    return k+1;  
  }
  
  /***
  UNSAT core lifting to generalize predecessor states
  t - predecessor cube
  inputs - values of input variables which lead to s
  p_inputs - values of primed input variables which may be implied by s
  ***/
  void QuipSolver::unsatCoreLift(const Cube & t, const Cube & inputs, const Cube & p_inputs, const Cube & s, Cube * lifted) const
  {
    logger.logorrheic() << "unsatCoreLift on cube "
                        << pc(t) 
                        << std::endl
                        << "s = " << pc(s) << std::endl
                        << "Before lifting: " << t.size() << std::endl;
 
    //do sat query t ^ inputs ^ p_inputs ^ T ^ ~s'
    vector<ID> assumps;
    assumps.insert(assumps.end(), inputs.begin(), inputs.end());
    assumps.insert(assumps.end(), p_inputs.begin(), p_inputs.end());
    assumps.insert(assumps.end(), t.begin(), t.end());
    
    Cube s_prime(s);
    for (size_t i=0; i<s.size(); i++){
      s_prime[i] = ev->prime(s[i]);
    }
    Clause not_s_prime;
    Expr::complement(*ev, s_prime, not_s_prime);
    
    //use different sat manager/view since we don't need any of the frame clauses here
    SAT::GID gid = lift->newGID();
    lift->add(not_s_prime, gid);
    stats.sat_calls++;
    bool sat = lift->sat(&assumps, NULL, &assumps);
    lift->remove(gid);
    assert(!sat);
    
    // It doesn't really make sense to call this without an output
    // Cube, but we'll leave the if(lifted) logic
    assert(lifted);
    if(lifted) {
      lifted->clear();
      ExprAttachment const * const eat = (ExprAttachment const *) model().constAttachment(Key::EXPR);
      for(vector<ID>::const_iterator it = assumps.begin(); it != assumps.end(); ++it) {
        if(eat->isStateVar(*it) || eat->isStateVar(ev->apply(Expr::Not, *it))) {
          lifted->push_back(*it);
        }
      }
      logger.logorrheic() << "UNSAT Core lifting: " << lifted->size() << std::endl; 
      logger.logorrheic() << "Lifted: " << pc(*lifted) << std::endl;
    }
  }
 
  /***
  Try to push every lemma in every frame as far as possible by repeatedly running
  consecution. As long as it passes consecution (i.e. is inductive), push it forward. 
  If a fully inductive frame Fk is found, where F_k = F_(k+1), push all its lemmas 
  to F_inf.
  Return Proof if F_inf => ~bad.
  ***/
  MC::ReturnType QuipSolver::propagateClauses()
  {
    logger.verbose() <<"\nStarting propagateClauses"<<endl;
    QuipAutoTimer auto_timer(stats.push_time);
    
    //record distribution of clauses
    stats.cube_dist_before_prop.push_back(vector<int>());
    for (size_t k = 1; k < frames.size(); k++){
      stats.cube_dist_before_prop.back().push_back(frames[k].size());
    }
    stats.cube_dist_before_prop.back().push_back(inf_cubes.size());
    stats.avg_cls_prop.push_back(0);
    
    for (int k=1; k < (int)frames.size(); k++){
      logger.logorrheic() << "k = " << k << std::endl;
      for (auto it = frames[k].begin(); it != frames[k].end(); it++){
        logger.logorrheic() << pc(*it) << std::endl;
      }
    }
    logger.logorrheic() << std::endl;
   
    for (int k=1; k < (int)frames.size() - 1; k++) {
      unsigned push_count = 0;
      if(!frames.back().empty()) {
        growFrames(frames.size() + 1);
      }
 
      int frame_sz = frames[k].size();

      // Copying the frame is not ideal, but it should be
      // be smallish and it simplifies things a lot by letting us
      // use addCube instead of directly modifying the inductive trace 
      // and the SAT solvers directly in here.
      // It would be even smaller if it stored CubeIDs instead of cubes
      CubeSet fk = frames[k];
      for (CubeSet::iterator it = fk.begin(); it != fk.end(); it++) {
        assert(isKnownLemma(*it));

        Lemma & lem = getLemma(*it);
        const Cube & c = lem.getCube();
        assert(c == *it);
        assert(lem.lv == k);

        if(requiresSGOnConsecution(opts.sg_type)) {
          assert(support_graph->getLevel(c) == k);
        }

        if (lem.bad) {
          //ignore bad cubes
          continue; 
        }
        
        //check if this cube can't be pushed due to unblocked CTP
        if (lem.ctp) {
          if (lem.ctp->blocked) {
            logger.logorrheic() << "Skipping pushing lemma " << pc(c) << " due to CTP" << endl;
            stats.ctp_cons_skipped++;
            continue;
          }
          else {
            lem.ctp->can_delete = true; //ctp will be deleted in the next call to checkCTPs
            lem.ctp = NULL;
          }
        }
        
        if (opts.sg_type != NO_SG) {
          //find the minimum level of all support cubes. If it's >= k then we can 
          //push this cube straight to level min_lv+1 without consecution calls 
          int min_lv = support_graph->getMinSupportLevel(c);
          if (min_lv >= k) {
            int next_lv = min_lv == INT_MAX ? INT_MAX : min_lv+1;
            logger.logorrheic() << "Pushing cube "
                                << pc(c) << " straight to level " << next_lv
                                << " due to supports" << endl; 
            addCube(c, next_lv);
            stats.supp_cons_skipped += (next_lv == INT_MAX ? (getLevel() - k) : (next_lv - k));
            push_count++;
            continue;
          }
        }
         
        SAT::Assignment asgn;
        initializeAsgn(asgn);
        stats.cons_from_prop_cls++; 
        bool do_sg = requiresSGOnPush(opts.sg_type) && !forbidsSGOnConsecution(opts.sg_type);
        if (consecution(c, INT_MAX, &asgn, NULL, do_sg)) {
          push_count++;
          addCube(c, INT_MAX);
        }
        else if (consecution(c, k, &asgn, NULL /* reduced */, do_sg)) { 
          push_count++;
          addCube(c, k+1);
        }
        else {
          if (opts.use_ctps) {
            Cube state;
            getCubesFromAsgn(asgn, &state);
            sort(state.begin(), state.end());
            CTP* ctp = new CTP(state);
            assert(k < (int)active_ctps.size());
            active_ctps[k].push_back(ctp);
            lem.ctp = ctp;
            logger.logorrheic() << "Adding CTP " << pc(state) 
                                << " for lemma " << pc(c) << endl;
            stats.num_ctps++;
          }
        }
      }
      
      if (frame_sz > 0) {
        stats.avg_cls_prop.back() += (double)push_count / (double)frame_sz;
      }
      
      if (frames[k].empty()) {
        logger.verbose() << "found inductive frame at k = " << k << endl;
        // Fk is fully inductive
        // push all cubes in F_k (i.e. frames[k+1]..frames[N-1]) to inf_cubes
        CubeSet inductive_frame;
        for (size_t i = k+1; i < frames.size(); i++) {
          CubeSet fi = frames[i];
          inductive_frame.insert(fi.begin(), fi.end());
          for(auto it = fi.begin(); it != fi.end(); ++it) {
            addCube(*it, INT_MAX);
          }
          assert(frames[i].empty());
        }
        break;
      }
    }
    
    //record distribution of clauses
    stats.cube_dist_after_prop.push_back(vector<int>());
    for (size_t k=1; k<frames.size(); k++){
      stats.cube_dist_after_prop.back().push_back(frames[k].size());
    }
    stats.cube_dist_after_prop.back().push_back(inf_cubes.size()); 
    stats.avg_cls_prop.back() /= getLevel(); //average over all levels
    
    //check if inf_cubes intersects bad
    //sat query F_inf ^ bad
    Expr::IDVector assumps;
    Cube bad = getCurrentBad();
    for (size_t i=0; i<bad.size(); i++){
      assumps.push_back(bad[i]);
    }
    stats.sat_calls++;
    bool is_sat = sat_view->sat(&assumps);
    if (!is_sat){
      logger.informative() << "Found inductive invariant" << endl;
      return MC::Proof;
    }
   
    return MC::Unknown;
  }

  void QuipSolver::constructSupportGraph(int level) {
    // There is no support graph at level 0
    assert(level > 0);
    assert(level == INT_MAX || (size_t)level < frames.size());

    logger.logorrheic() << "Constructing support graph for level " 
                        << (level == INT_MAX ? "infinity" : std::to_string(level))
                        << std::endl;

    const CubeSet & f_i = level == INT_MAX ? inf_cubes : frames[level];

    for(auto it = f_i.begin(); it != f_i.end(); ++it) {
      const Cube & cube = *it;
      assert(isKnownLemma(cube));
      assert(getLemma(cube).lv == level);
      getLemma(cube).annotation = stringOfClauseOrCube(cube);
      computeSupport(cube);
    }
  }
   
  /***
   Build the support graph from scratch.
   ***/
  void QuipSolver::constructSupportGraph() {
    assert(support_graph);
    delete support_graph;
    support_graph = new SupportGraph(*cube_db);

    // true => include activation literals for infinity lemmas
    renewSAT(true);

    for(size_t i = 1; i < frames.size(); ++i) {
      constructSupportGraph(i);
    }
    constructSupportGraph(INT_MAX);
  }
  
  /***
  Write support graph to dot file. sg_type option specifies which
  graph is written.
  ***/
  void QuipSolver::writeSupportGraph() const {
    if(opts.sg_infdot_file != "") {
      support_graph->writeInfDot(opts.sg_infdot_file, getCurrentBad());
    }

    if(opts.sg_dot_file != "") {
      support_graph->writeDot(opts.sg_dot_file, getCurrentBad());
    }
  }
  
  /***
  If s is successfully blocked, returns (t,g) where t is a lemma (cube) such that 
  !t blocks s, and g is the level to which that !t holds.
  If s cannot be blocked, returns a predecessor cube of s and g = k-1.
  ***/
  pair<Cube, int> QuipSolver::block(const ProofObligation & po, Cube * inputs, Cube * concrete_state)
  {
    const Cube & s = po.cube;
    int k = po.level;
    logger.logorrheic() << "block, level "<<k<<", cube " << pc(s) << std::endl;
    assert(k <= getLevel());

    // Syntactically check if s is blocked by a cube in one of the frames    
    for (auto it = inf_cubes.begin(); it != inf_cubes.end(); it++){
      if (subsumes(s, *it)){
        logger.logorrheic() << "syntactic check returning inf" << endl;
        stats.syntactic_block_success++;
        return make_pair(*it, INT_MAX);
      }
    }
    
    for (int i = frames.size() - 1; i >= k; i--) {
      for (auto it = frames[i].begin(); it != frames[i].end(); it++) {
        if (subsumes(s, *it)) {
          logger.logorrheic() << "syntactic check returning " << i << endl;
          stats.syntactic_block_success++;
          // Note: it's important to add the cube, even though it's subsumed.
          // The subsuming lemma might be a bad lemma, and this one might not.
          // We don't do it in the case of F_inf because obviously the 
          // subsuming lemma is good and not bad
          if (s != *it) {
            addCube(s, i);
          }
          return make_pair(*it, i);
        }
      }
    }
    stats.syntactic_block_fail++;
    
    //check satisfiability of F_(k-1) ^ T ^ s'    
    SAT::Assignment asgn;
    // include primed state
    bool include_pstate = opts.print_cex;
    initializeAsgn(asgn, include_pstate);
    
    bool cons = consecution(s, k-1, &asgn);
    logger.logorrheic() << "cons: " << cons << endl;
    
    if (!cons) {
      // Extract and return predecessor state t and optionally concrete_state 
      // (the concrete s-state involved in the assignment)
      //pinp = primed inputs (forced by s)
      Cube t, lifted_t, local_inputs, pinp;       
      Cube & inp = inputs ? *inputs : local_inputs;

      getCubesFromAsgn(asgn, &t, &inp, &pinp, 
                       include_pstate ? concrete_state : NULL);
      sort(t.begin(), t.end());
      //generalize the predecessor
      if(opts.lift) {
        unsatCoreLift(t, inp, pinp, s, &lifted_t);
        t = lifted_t;
      }
      return make_pair(t, k-1);
    }
    else {
      Cube t(s);
      //generalize t
      int g = k;
      if (!po.from_sg) {
        g = generalize(k-1, frames.size()+1, t);
      }
      logger.logorrheic() << "generalize returned g = " << g 
                          << ", cube = " << pc(t) 
                          << std::endl;
      stats.cube_size_total += (int)t.size();
      stats.cube_size_cnt++;
      assert(g == INT_MAX || g <= (int)frames.size());
      addCube(t,g);
      if(requiresSGOnPush(opts.sg_type)) {
        computeSupport(t);
      }
      return make_pair(t, g);
    }
  }
  
  /***
  Return a state (cube) r such that r is 1-step reachable from r0 and
  intersects the parent cube of po. 
  ***/
  ReachableState * QuipSolver::extractReachable(ProofObligation * po, const Cube & r0)
  {
    /*
    "Simulate" the transition r0->r with the sat query r0 ^ T ^ r', where r' is 
    a cube of all the state variables. All variables in the formula are assigned in 
    the assumptions except the literals not in po->parent.
    */
    assert(po->parent != NULL);
    logger.logorrheic() << "extracting reachable state, r0 = " << pc(r0) << std::endl;
    Expr::IDVector assumps;
    //assumps.push_back(levelVar(po->level));
    
    //assert r0 in the assumptions
    for (size_t i=0; i<r0.size(); i++){
      assumps.push_back(r0[i]);
    }
    
    //assert the input values
    for (size_t i=0; i<po->inputs.size(); i++){
      assumps.push_back(po->inputs[i]);
    }    
    
    //make sure r intersects po->parent by asserting values of the literals in po->parent
    for (size_t i=0; i<po->parent->cube.size(); i++){
      assumps.push_back(ev->prime(po->parent->cube[i]));
    }
    
    SAT::Assignment asgn;
    //interested in all the primed state variables
    for (size_t i=0; i<state_vars.size(); i++){
      asgn.insert(SAT::Assignment::value_type(ev->prime(state_vars[i]), SAT::Unknown));
    }
    
    stats.sat_calls++;
    bool is_sat = sat_view->sat(&assumps, &asgn);
    assert(is_sat);
    
    Cube r;
    for (SAT::Assignment::iterator it = asgn.begin(); it!=asgn.end(); it++){
      assert(it->second != SAT::Unknown);
      r.push_back(ev->unprime(it->second == SAT::False ? 
        ev->apply(Expr::Not, it->first) : 
        it->first));
    }
    sort(r.begin(), r.end());
    
    logger.logorrheic() << "returning r = " << pc(r) << std::endl;
    
    //mark all cubes excluding r as bad
    for (size_t i = 1; i < frames.size(); i++){
      for (auto it = frames[i].begin(); it != frames[i].end(); it++){
        assert(isKnownLemma(*it));
        Lemma & lem = getLemma(*it);
        const Cube & cube = lem.getCube();
        if (cube != getCurrentBad() && !lem.bad && subsumes(r, cube)) {
          logger.logorrheic() << "Marking as bad " << pc(cube) << std::endl;
          markBad(cube);
        }
      }
    }
    
    Cube inputs = po->inputs;
    ReachableState * rs = rsman.addState(r, inputs);
    return rs;
  }
 
  ReturnValue QuipSolver::buildCex(ProofObligation & obl, ReachableState * r0) const {
    ReturnValue rv = ReturnValue(MC::CEX);

    //Construct counterexample.
    //Downwards through reachable states to initial.
    vector<Transition> tmp;
    while (r0) {
      if (r0->pre)
        tmp.push_back(Transition(r0->pre->state, r0->inputs));
      else
        tmp.push_back(Transition(init_cube, r0->inputs));
      r0 = r0->pre;
    }

    //add to rv.cex in reverse order
    for(auto it = tmp.rbegin(); it != tmp.rend(); ++it) {
      rv.cex.push_back(*it);
    }

    ProofObligation * po = &obl;
    while(po) {
      rv.cex.push_back(Transition(po->cex_state, po->inputs));
      po = po->parent;
    }
    return rv;
  }

  /***
  Handle a failed may proof obligation. Extract and store the trace, and mark 
  the obligations in the chain as failed. Returns a ReachableState * for the
  first reachable state in the trace.
  ***/
  ReachableState * QuipSolver::handleFailedMayProof(ProofObligation * po, const Cube & pred) {
    assert(po->type == MAY_PROOF);

    logger.logorrheic() << "Failed may proof: " << pc(po->cube) << std::endl;

    ReachableState * r0 = extractReachable(po, pred);

    // Go up through the chain of obligations and mark them as failed.
    // Otherwise, we may end up in a situation where we drop the 
    // reachable state we just learned (or if max_reachable_states is 0,
    // never store it) and therefore reconsider this obligation repeatedly

    if(isKnownLemma(po->cube)) {
      markBad(po->cube);
    }

    po = po->parent;
    ReachableState * r = r0;
    while (po) {
      assert(po->type == MAY_PROOF);
      po->failed = true;

      if(isKnownLemma(po->cube)) {
        markBad(po->cube);
      }

      if(po->parent) {
        ReachableState * r_new = extractReachable(po, r->state); 
        r_new->pre = r;
        r = r_new;
      }
      po = po->parent;
    }

    return r0;
  }

  void QuipSolver::logObligation(const ProofObligation * po) const {
    size_t ss_size = 0;
    if (isKnownLemma(po->cube)) {
      ss_size = support_graph->getInDegree(po->cube);
    }
    
    logger.logorrheic() << (po->type == MUST_PROOF ? "Must" : "May")
                        << " proof obligation at level " << po->level
                        << " and depth " << po->depth
                        << " for cube " << pc(po->cube) 
                        << ", effort "<< po->effort
                        << ", limit " << po->effort_limit
                        << ", in-degree " 
                        << (ss_size > 0 ? std::to_string(ss_size) : "unknown")
                        << std::endl;
  }

  /***
  Try to block bad_cube at level k.
  ***/
  ReturnValue QuipSolver::blockCube(const Cube & bad_cube, int k, int effort_limit)
  {
    ObligationPool pool;
    std::list<ProofObligation *> sg_obligations;
    ObligationQueue q;
    ProofObligation * po = newObligation(pool, bad_cube, k, MUST_PROOF, NULL, effort_limit);
    q.push(po);
    stats.recordStat(stats.must_proof_dist, k);
    
    while (!q.empty()) {
      po = q.top();
      assert(!po->from_sg || po->type == MAY_PROOF);
      q.pop();

      if (po->failed) {
        assert(po->type == MAY_PROOF);
        stats.may_fail++;
        deactivateProofObligation(po);
        continue;
      }

      logObligation(po);

      // Time the obligation in the appropriate bucket
      // We start the timer here so we don't count obligations that have 
      // already failed
      double & time_bucket = (po->type == MUST_PROOF) ? stats.must_time : stats.may_time;
      QuipAutoTimer auto_timer(time_bucket);

      // Counter-example found the old-fashioned way 
      // (proof obligations going back to 0)
      if (po->level == 0) {
        if (po->type == MUST_PROOF) {
          po->cex_state = init_cube;
          return buildCex(*po);
        } else {
          // Handle the failure, generating a trace of reachable states starting
          // from the initial state
          stats.may_fail++;
          handleFailedMayProof(po, init_cube);

          if(po->from_sg) {
            for (ProofObligation * sgo : sg_obligations) {
              sgo->failed = true;
            }
            sg_obligations.clear();
          }

          deactivateProofObligation(po);
          continue;
        }
      }
     
      // Note that Truss/QuipSS/QuipSG has its own effort limit
      // sg_effort. These should be unified, but they're not
      if (po->effort > po->effort_limit) {
        //This should mean that all ancestors have also had their effort limits 
        //exceeded (when summed over all descendents) so we abandon them all. 
        logger.informative() << "Abandoning proof obligation due to effort limit exceeded." << endl;
        return ReturnValue(MC::Unknown);
      }
      
      ReachableState * r0 = rsman.intersect(po->cube);      
      // Counter-example found via reachable states
      if (r0) {
        logger.logorrheic() << "Cube intersects reachable state " 
                            << pc(r0->state) 
                            << std::endl;
        if (po->type == MUST_PROOF) {
          // buildCex ignores r0's state, so we can use it in po
          po->cex_state = r0->state;
          return buildCex(*po, r0);
        }
        else {
          if (po->parent == NULL) {
            // Just mark the cube as bad if it's a known lemma
            if (isKnownLemma(po->cube)) {
              markBad(po->cube);
            }
          } else {
            // Since po->cube is known reachable, generate a trace starting from
            // r0->state.
            ReachableState * r = handleFailedMayProof(po, r0->state);
            if(po->from_sg) {
              for (ProofObligation * sgo : sg_obligations) {
                sgo->failed = true;
              }
              sg_obligations.clear();
            }

            r->pre = r0;
          }
          stats.may_fail++;
          deactivateProofObligation(po);
          continue;
        }
      }

      // For EXPERIMENTAL_SG, first try to meet the obligation using the 
      // support graph - This is Truss/QuipSS/QuipSG
      // Note: this procedure is more complicated than necessary to implement
      // Truss. It supports a lot of other options.

      // Option to mix support graph with standard strategy - SG on even 
      // levels, standard on odd levels
      bool option_skip = (opts.sg_even_only && k % 2 != 0);

      if (opts.sg_type == EXPERIMENTAL_SG && isKnownLemma(po->cube) && !option_skip) 
      {
        const Cube & cube = po->cube;
        int support_lvl = getSupportLevel(cube);
        if(support_lvl >= opts.min_sg_level) {
          bool acceptable_known = hasAcceptableSupport(cube);
          if (!acceptable_known) {
            support_graph->deleteSupport(cube);
            logger.logorrheic() << "Support set unnacceptable, trying to find new one for" 
                                << pc(cube)
                                << std::endl;
            acceptable_known = computeAcceptableSupport(cube);
          } 

          if (!acceptable_known && opts.desperate_sg) { 
            support_graph->deleteSupport(cube);
            logger.logorrheic() << "Support set unnacceptable, desperately trying to find new one for" 
                                << pc(cube)
                                << std::endl;
            acceptable_known = computeAcceptableSupport(cube, opts.min_sg_level);
          }

          if (acceptable_known) {
            // Enqueue the support, re-enqeue po, continue
            size_t enqueued = 0;
            auto support_range = support_graph->getInEdges(cube);
            int min_lvl = INT_MAX;
            for (auto it = support_range.first; it != support_range.second; ++it) {
              CubeID id = *it;
              const Cube & in_cube = cube_db->getCube(id);
              const Lemma & in_lem = getLemma(in_cube);
              int lvl = in_lem.lv;
              min_lvl = std::min(min_lvl, in_lem.lv);
              if (lvl < po->level - 1) {
                logger.logorrheic() << "Enqueuing " << pc(in_cube) 
                                    << " from level " << lvl 
                                    << " to level " << po->level - 1
                                    << std::endl;
                ProofObligation * new_po = newObligation(pool, in_cube, po->level - 1, MAY_PROOF, NULL);
                new_po->from_sg = true;
                sg_obligations.push_back(new_po);
                q.push(new_po);
                enqueued++;
              }
            }

            logger.logorrheic() << "Acceptable support set found with " 
                                << support_graph->getInDegree(cube)
                                << " lemmas (" << enqueued << " enqueued)"
                                << std::endl;

            // There's no guarantee anything will be enqueued, but if nothing
            // is, it should mean the obligation will pass
            if (enqueued > 0) {
              q.push(po);
              continue;
            } else {
              assert(min_lvl >= po->level - 1);
              int to_lvl = min_lvl;
              if (to_lvl < INT_MAX) {
                to_lvl++;
              }
              logger.logorrheic() << "Sending " << pc(cube) 
                                  << " straight to level " << to_lvl 
                                  << std::endl;
              addCube(cube, to_lvl);
              continue;
            }
          } else {
            logger.logorrheic() << "Unable to find acceptable support set" << std::endl;
          }
        } else {
          logger.logorrheic() << "Not seeking support set due to level "
                              << "(supported at: " << support_lvl << ")"
                              << std::endl;
        }
      } 

      if (opts.sg_type == EXPERIMENTAL_SG && po->from_sg) {
        po->sg_effort++;
        // Skip because of effort - If we're below min_sg_level, then we use a 
        // different limit than we'd use for higher levels. This is because
        // the levels below min_sg_level can't possibly have a support set
        int support_lvl = getSupportLevel(po->cube);
        int limit = (support_lvl < opts.min_sg_level) ? 
                             opts.sg_effort_limit_minlvl : 
                             opts.sg_effort_limit;

        if (po->sg_effort > limit) {
          logger.logorrheic() << "Cube " << pc(po->cube) 
                              << " marked ugly due to effort (used: "
                              << po->sg_effort - 1 
                              << " / " << limit << ")" 
                              << std::endl;
          markUgly(po->cube);

          for (ProofObligation * sgo : sg_obligations) {
            sgo->failed = true;
          }
          sg_obligations.clear();

          continue;
        }
      }
     
      //very rough measure for now. block should be far more expensive than
      //anything else in this function, so only count effort if it gets here.
      po->effort++; 
      stats.recordStat(stats.effort_by_level, po->level);
      
      Cube inputs, cex_state;
      Cube t; 
      int g; 
      std::tie(t, g) = block(*po, &inputs, &cex_state); 
      
      logger.logorrheic() << "block returned g = " << g << " (k = " << k << ")" 
                          << ", t = " << pc(t) 
                          << ", cex_state = " << pc(cex_state) 
                          << std::endl;

      if (g != po->level-1) {
        //po->cube is blocked by !t
        if (po->type == MAY_PROOF) {
          stats.may_success++;
        } else {
          stats.must_success++;
        }

        if (g < INT_MAX && (g + 1) < k) {
          if(opts.reenqueue) {
            //reinsert proof obligation for t to try to push it as far as possible 
            po->level = g + 1;
            if (t != po->cube) {
              // It's important for some unsafe instances to re-enqueue both obligations.
              ProofObligation *copy = newObligation(pool, *po);
              stats.recordStat(copy->type == MAY_PROOF ? stats.may_proof_dist : stats.must_proof_dist, copy->level);
              q.push(copy);

              // TODO: this should have it's own separate effort 
              po->type = MAY_PROOF;
              po->parent = NULL;  
              po->cube = t;
              po->effort = 0;
            }
            q.push(po);
          }
          stats.recordStat(po->type == MAY_PROOF ? stats.may_proof_dist : stats.must_proof_dist, po->level);
        }
        else {
          stats.recordStat(stats.effort_dist, po->effort);
          deactivateProofObligation(po);
        }
      }
      else {
        //t is a predecessor of po->cube
        po->cex_state = cex_state;
        //effort set automatically in newObligation
        ProofObligation * next_po = newObligation(pool, t, po->level-1, po->type, po);
        next_po->inputs = inputs;
        q.push(next_po); 
        stats.recordStat(next_po->type == MAY_PROOF ? stats.may_proof_dist : stats.must_proof_dist, next_po->level);
        q.push(po);
      }
    }
   
    // Probably better to return a Boolean
    return ReturnValue(MC::Proof); //blocked
  }
 
  ReturnValue QuipSolver::infBlockCube() { 
    return infBlockCube(getCurrentBad());
  }

  /***
  Block the cube at level infinity.
  ***/
  ReturnValue QuipSolver::infBlockCube(const Cube & bad_cube)
  {
    ReturnValue rv = ReturnValue(MC::Unknown);
    logger.informative() << "Trying to block cube " << pc(bad_cube) << std::endl;
    
    if (checkInitial(bad_cube)) {
      rv.returnType = MC::CEX;
      //store empty vectors in cex since the trace is just the initial state
      Cube inputs;
      rv.cex.push_back(Transition(init_cube, inputs));
      return rv;
    }
   
    // Initial states are not bad, so add the cube here  
    addCube(bad_cube, 0);

    //Main quip loop - try to strengthen the inductive set until bad_cube is blocked.
    setLevel(0);
    while (true) {
      incrementLevel();

      const Lemma & bad_lemma = getLemma(getCurrentBad());
      while(bad_lemma.lv >= getLevel() && bad_lemma.lv < INT_MAX) {
        incrementLevel();
      }

      int k = getLevel();

      logger.terse() << "\nLevel "<< k << endl;
      stats.level = max(stats.level, k); //use max in case infBlockCube is called again

      ReturnValue block_rv = blockCube(bad_cube, k);
      assert(block_rv.returnType != MC::Unknown);
      
      if (block_rv.returnType == MC::CEX) {
        logger.informative() << "Counterexample of length " 
                             << block_rv.cex.size()
                             << " found at level " << k << endl;
        return block_rv;
      }

      if (propagateClauses() == MC::Proof) {
        rv.returnType = MC::Proof;
        // set inf_cubes as proof
        for (auto it = inf_cubes.begin(); it != inf_cubes.end(); it++) {
          Clause cls;
          const Cube & cube = *it;
          if(cube != getCurrentBad()) {
            Expr::complement(*ev, cube, cls);
            rv.proof.push_back(cls);
          }
        }
        return rv;
      } 

      if (opts.renew_sat) {
        renewSAT();
      }
    }
    
    return rv;
  }

  void QuipSolver::initTR() {
    assert(tr_clauses.empty());
    ID npi = getNPI();

    //get the clauses for the transition relation 
    if(opts.iictl_clauses) {
      getIICTLClauses(tr_clauses, tr_clauses, npi);
    } else {
      assert(!opts.constraints);
      assert(!opts.neg_constraints);
      getTRClauses(&tr_clauses, &tr_clauses, npi, true);
    }
  }

  /***
  Get the clauses for the transition relation (in cons) and the primed
  versions (in error). Both cons and error are nullable.
  ***/
  void QuipSolver::getTRClauses(vector<Clause> * cons, vector<Clause> * error, ID npi, bool silent) const {
    // Note that cons and error could be the same pointer
    assert(!cons || cons->empty());
    assert(!error || error->empty());
    CNFAttachment * cnfat = (CNFAttachment *) model().constAttachment(Key::CNF);
    ExprAttachment * eat = (ExprAttachment *) model().constAttachment(Key::EXPR);

    Cube bad = getCurrentBad();
    // TODO: fix handling of bad. The way IC3 allows multiple calls is by 
    // changing the output of the model. I think that's what we'll have to
    // do here as well.
    assert(bad.size() == 1);

    if(cons) {
      vector<Clause> & cons_clauses = *cons;
      cons_clauses = cnfat->getCNF();
      assert(cons_clauses.back().size() == 1);
      cons_clauses.back().push_back(bad[0]);
      cons_clauses.push_back(Clause());
      Expr::complement(*ev, cons_clauses[(int)cons_clauses.size()-2], cons_clauses.back());
      logger.logorrheic(!silent) << "cons_clauses:" << endl; 
      for (size_t i = 0; i < cons_clauses.size(); i++) {
        sort(cons_clauses[i].begin(), cons_clauses[i].end());
        logger.logorrheic(!silent) << pc(cons_clauses[i]) << std::endl; 
      }

      // The constraints are just a vector of vectors of clauses. 
      // There is no difference between constraints[0] and constraints[1], etc.
      if(opts.constraints) {
        for(auto it = opts.constraints->begin(); it != opts.constraints->end(); ++it) {
          cons_clauses.insert(cons_clauses.end(), it->begin(), it->end());
        }
      }
    }

    if(error) {
      vector<Clause> & error_clauses = *error;
      Expr::tseitin(*ev, Expr::primeFormula(*ev, npi), error_clauses);
      assert(error_clauses.back().size() == 1);
      error_clauses.back().push_back(ev->apply(Expr::Not, ev->prime(bad[0])));
      error_clauses.push_back(Clause());
      Expr::complement(*ev, error_clauses[(int)error_clauses.size()-2], error_clauses.back());
      logger.logorrheic(!silent) << "error_clauses:" << endl;
      for (size_t i=0; i<error_clauses.size(); i++){
        sort(error_clauses[i].begin(), error_clauses[i].end());
        logger.logorrheic(!silent) << pc(error_clauses[i]) << std::endl;
      } 
    }

    model().constRelease(eat);
    model().constRelease(cnfat);
  }

  /***
  This is more like a stub, in the future we should support multiple
  queries with different bad states.
  ***/
  Cube QuipSolver::getCurrentBad() const {
    Cube bad;
    bad.push_back(levelVar(-1));
    return bad;
  }

  void QuipSolver::getIICTLClauses(std::vector<Clause> & cons_clauses, std::vector<Clause> & error_clauses, ID npi) {
    assert(opts.iictl_clauses);
    // TODO: handle neg constraints
    // IICTL still passes them with an empty clause
    if(opts.neg_constraints) {
      const SAT::Clauses & clauses = *opts.neg_constraints;
      for(auto it = clauses.begin(); it != clauses.end(); ++it) {
        const Clause & c = *it;
        assert(c.empty());
      }
    }

    getTRClauses(&cons_clauses, NULL, npi, true);

    // IICTL gives a CNF representation of the property in iictl_clauses
    // But IICTL isn't supported right now anyway
    Cube bad = getCurrentBad();
    for(auto clauses_it = opts.iictl_clauses->begin(); 
             clauses_it != opts.iictl_clauses->end(); 
             ++clauses_it) {
      vector<ID> cls;
      for(auto it = clauses_it->begin(); it != clauses_it->end(); ++it) {
        ID lit = *it;
        if(lit != ev->btrue() && lit != ev->bfalse()) {
          cls.push_back(ev->prime(lit));
        } else {
          cls.push_back(lit);
        }
      }
      error_clauses.push_back(cls);
    }
    assert(error_clauses.back().size() == 1);
    error_clauses.back().push_back(ev->apply(Expr::Not, ev->prime(bad[0])));
    error_clauses.push_back(Clause());
    Expr::complement(*ev, error_clauses[(int)error_clauses.size()-2], error_clauses.back());
    for (size_t i=0; i<error_clauses.size(); i++){
      sort(error_clauses[i].begin(), error_clauses[i].end());
    } 
  }

  /***
  Load the transition relation into the given manager. 
  Returns true if a SAT::Trivial is thrown, false otherwise
  ***/
  bool QuipSolver::loadTR(SAT::Manager * manager) {
    assert(manager);
  
    //add cons_clauses to the sat manager
    bool trivial = false;
    try {
      manager->add(tr_clauses);
    }
    catch (SAT::Trivial const & tr) {
      assert(!tr.value());
      //case where initial states intersects with bad states
      logger.informative() << "The property fails trivially." << endl;
      trivial = true;
    }

    return trivial;
  }

  /***
  Loads the lemmas from every frame (including infinity) into the manager
  given with appropriate activation literals
  ***/
  void QuipSolver::loadFrames(SAT::Manager * manager) {
    CubeSet frameCubes;
    for(size_t i = frames.size(); i > 0; --i) {
      const CubeSet & cubes = frames.at(i-1);
      frameCubes.insert(cubes.begin(), cubes.end());
      addCubeSetToManager(frameCubes, i-1, manager);
    }

    // Load infinity lemmas
    addCubeSetToManager(inf_cubes, INT_MAX, manager); 
  }

  /***
  Loads the lemmas from every frame (including infinity) into the manager
  with individual activation literals per lemma
  ***/
  void QuipSolver::loadFramesWithSGLits(SAT::Manager * manager, bool inf_act) {
    for(size_t i = 0; i < frames.size(); ++i) {
      const CubeSet & cubes = frames.at(i);
      for(auto cube_it = cubes.begin(); cube_it != cubes.end(); cube_it++) {
        const Cube & cube = *cube_it;
        assert(isKnownLemma(cube));
        addCubeToSGManager(cube, true, manager);
      }
    }

    for(auto cube_it = inf_cubes.begin(); cube_it != inf_cubes.end(); cube_it++) {
      const Cube & cube = *cube_it;
      assert(isKnownLemma(cube));
      // If inf_act is true, include activation literals
      addCubeToSGManager(cube, inf_act, manager);
    }
  }

  /***
  Initializes the solver used in lifting
  ***/
  void QuipSolver::initLiftingSAT() {
    assert(lift_man);
    if(lift) {
      delete lift;
      lift = NULL;
    }
    std::string backend = options().backend;
    lift = lift_man->newView(*ev, backend);
  }

  /***
  Initializes the primary SAT solver used in most queries
  ***/
  void QuipSolver::initSAT() {
    assert(sat_man);
    if(sat_view) {
      delete sat_view;
      sat_view = NULL;
    }
    std::string backend = options().backend;
    //enable trace generation only for unsat core support graph
    bool trace_gen = isCoreBasedSG(opts.sg_type);
    sat_view = sat_man->newView(*ev, backend, trace_gen); 
  }

  /***
  Initializes the SAT solver used in support graph queries
  ***/
  void QuipSolver::initSGSAT() {
    assert(sg_man);
    if(sg_view) {
      delete sg_view;
      sg_view = NULL;
    }
    std::string backend = options().backend;
    bool trace_gen = false;
    sg_view = sg_man->newView(*ev, backend, trace_gen); 

    if(opts.abbr_backend) {
      sg_view->lockInModel();
    }
  }

  /***
  Returns a pointer to an initialized SAT Manager
  ***/
  SAT::Manager * QuipSolver::initSATManager(bool load_frames, bool load_inf) {
    SAT::Manager * manager = model().newSATManager();

    // Returns true if the SAT problem is trivial
    if(loadTR(manager)) {
      delete manager;
      return NULL;
    }

    if(load_frames) {
      loadFrames(manager);
    } else if(load_inf) {
      addCubeSetToManager(inf_cubes, INT_MAX, manager); 
    }

    return manager;
  }

  /***
  Sets up F_0 from the model's initial conditions
  ***/
  void QuipSolver::initFrames() {
    for (size_t i = 0; i < init_cube.size(); i++) {
      Cube cube;
      cube.push_back(ev->apply(Expr::Not, init_cube[i]));
      addCubeToTrace(cube, 0); 
    }

    rsman = ReachableStateManager(opts.max_reachable_states);
    rsman.addState(init_cube, Cube()); 
  }

  /***
  (Re-) initializes the solvers.
  Returns true if initSATManager gives back a NULL, which suggests that
  the problem is trivially SAT. The first time loading up clauses, this
  can mean that the property fails trivially. Later on, it probably
  means something went wrong.
  Loads infinity lemmas in the SG manager iff sg_man_inf = true
  ***/
  bool QuipSolver::renewSAT(bool sg_man_inf) {
    logger.verbose() << "renewSAT" << std::endl;

    cleanupSAT();

    sat_man = initSATManager();
    if(sat_man == NULL) {
      return true;
    }

    // With infinity lemmas, without other frames
    lift_man = initSATManager(false, true);
    // If the problem were trivial, I think sat_man should've
    // already came back as NULL above. So lift_man == NULL probably
    // implies an error
    assert(lift_man != NULL);

    // Support graph manager with no lemmas (it needs them loaded specially)
    sg_man = initSATManager(false, false);
    assert(sg_man != NULL);
    loadFramesWithSGLits(sg_man, sg_man_inf);
    sg_man_has_inf_act = sg_man_inf;

    initLiftingSAT();
    initSAT();
    initSGSAT();
    return false;
  }

  /***
  Sets up and initializes clauses/cubes from the underlying model. 
  May return a value indicating success or failure if the problem is trivial.
  ***/
  ReturnValue QuipSolver::quipInit()
  {
    logger.informative() << "Starting quipInit" << endl;

    ExprAttachment * eat = (ExprAttachment *) model().constAttachment(Key::EXPR);

    ReturnValue trivial_cexrv(MC::CEX);
    trivial_cexrv.cex.push_back(Transition(init_cube, Cube()));

    ID npi = eat->outputFnOf(eat->outputs()[0]);
    if (npi == ev->bfalse()) {
      model().constRelease(eat);
      return ReturnValue(MC::Proof);
    }
    if (npi == ev->btrue()) {
      model().constRelease(eat);
      return trivial_cexrv;
    }

    // Set up the SAT managers and solvers
    bool trivial = renewSAT();
    if(trivial) {
      logger.informative() << "The property fails trivially." << std::endl;
      model().constRelease(eat);
      return trivial_cexrv;
    }
 
    model().constRelease(eat);
    
    return ReturnValue(MC::Unknown);
  }

  void QuipSolver::printStats() {
    stats.proof_size = inf_cubes.size();
    for (unsigned k = 1; k < frames.size(); k++) {
      for (auto it = frames[k].begin(); it != frames[k].end(); it++) {
        const Cube & cube = *it;
        assert(isKnownLemma(cube));
        const Lemma & lem = getLemma(cube);

        if (lem.bad) {
          stats.bad_lemmas++;
        }

        if (lem.ugly) {
          stats.ugly_lemmas++;
        }
        
        stats.total_lemmas++;
      }
    }

    stats.total_lemmas += inf_cubes.size();

    // TODO The stats should all be the QuipStats class,
    // not spread across all of these classes.
    stats.print();
    rsman.printStats();

    if (opts.sg_type != NO_SG) {
      support_graph->computeStats(getCurrentBad(), false);
      support_graph->printStats(false);
      support_graph->computeStats(getCurrentBad(), true);
      support_graph->printStats(true);
    }
  }
  
  void ReachableStateManager::printStats() const
  {
    cout << "Number of reachable states found: " << total_states << endl;
    cout << "Number of proof obligations blocked due to reachable state intersections: " << num_intersections << endl;
    cout << "Number of reachable state evictions: " << num_evictions << endl;
  }
  
  void QuipStats::recordStat(vector<int> & stat, int x) {
    while ((int)stat.size() <= x) 
      stat.push_back(0);
    stat[x]++;
  }
  
  string QuipStats::str(const vector<int> & stat) const {
    std::stringstream ss;
    for (unsigned i=0; i<stat.size(); i++){
      ss<<" "<<stat[i];
    }
    return ss.str();
  }
  
  void QuipStats::print() const
  {
    cout << "============================ Quip Statistics ============================" << endl;
    cout << "Final level reached: " << level << endl;
    cout << "Number of initial states: " << init_states << endl;
    cout << "Proof size: " << proof_size << endl;
    cout << "Number of consecution calls: " << consecution_calls << endl;
    cout << "Number of consecution calls from propagateClauses only: " << cons_from_prop_cls << endl;
    cout << "Number of sat queries: " << sat_calls << endl;
    cout << "Number of cubes blocked by syntactic check: " << syntactic_block_success << endl;
    cout << "Number of cubes not blocked by syntactic check: " << syntactic_block_fail << endl;
    
    //Note: last output number is always F_inf size
    for (unsigned k = 0; k < cube_dist_before_prop.size(); k++) {
      cout << "Cube distribution before pushing at level " << k+1 << " :";
      //don't print out a whole bunch of zero frames
      int last_nonzero = (int) cube_dist_before_prop[k].size() - 2; //always include last (F_inf)
      for (; last_nonzero > 0; last_nonzero--) {
        if (cube_dist_before_prop[k][last_nonzero] > 0) {
          break;
        }
      }

      for (int i = 0; i <= last_nonzero; i++) {
        cout << " " << cube_dist_before_prop[k][i];
      }
      cout << " ... " << cube_dist_before_prop[k].back() << endl; //F_inf
      
      cout << "Cube distribution after pushing at level  " << k+1 << " :";
      last_nonzero = cube_dist_before_prop[k].size() - 2;
      for (; last_nonzero > 0; last_nonzero--) {
        if (cube_dist_after_prop[k][last_nonzero] > 0) {
          break;
        }
      }
      for (int i = 0; i <= last_nonzero; i++) {
        cout << " " << cube_dist_after_prop[k][i];
      }
      cout << " ... " << cube_dist_after_prop[k].back() << endl; //F_inf
    }
    
    for (unsigned k=0; k<avg_cls_prop.size(); k++) {
      cout << "Average percentage of cubes pushed at level " << k+1 << " : " << avg_cls_prop[k]*100 << "%" << endl;
    }
    
    cout << "Average size of blocking cubes after generalization: " << (double)cube_size_total/cube_size_cnt << endl;
    cout << "Average size reduction of cubes during consecution: " << cube_reduction_total/cube_reduction_cnt << endl;
    
    if (core_cnt > 0) {
      //unsat core stats
      cout << "Average size of unsat core: " << (double)core_size_total/core_cnt << endl;
      cout << "Average of (unsat core size)/(F_k + F_inf size): " << core_vs_frame_size/core_cnt << endl;
    }
    cout << "Number of lemmas: " << total_lemmas << endl;
    cout << "Number of bad lemmas: " << bad_lemmas << endl;
    cout << "Number of ugly lemmas: " << ugly_lemmas << endl;
    cout << "Number of times lemmas were marked ugly: " << ugly_marked << endl;
    cout << "Number of times lemmas were unmarked ugly: " << ugly_unmarked << endl;
    
    cout << "Number of CTPs found: " << num_ctps << endl;
    cout << "Number of consecution calls skipped due to CTPs: " << ctp_cons_skipped << endl;
    cout << "Number of CTP subsumption checks: " << ctp_subsummption_checks << endl;
    cout << "Number of consecution calls skipped due to support lemmas: " << supp_cons_skipped << endl;
  
    cout << "Successful must proof obligations: " << must_success << endl;
    cout << "Successful may proof obligations: " << may_success << endl;
    cout << "Failed may proof obligations: " << may_fail << endl;
    cout << "Support set computations: " << support_calls << endl;
    cout << "Must proof obligation distribution by level:" << str(must_proof_dist) << endl;
    cout << "May proof obligation distribution by level:" << str(may_proof_dist) << endl;
    cout << "Total effort on proof obligations by level:" << str(effort_by_level) << endl;
    cout << "Effort histogram for successful proof obligation: " << str(effort_dist) << endl;
    
    cout << "Time spent in Quip: " << quip_time << endl; 
    cout << "Time spent processing must-proofs: " << must_time << endl; 
    cout << "Time spent processing may-proofs: " << may_time << endl; 
    cout << "Time spent computing support sets: " << support_time << endl; 
    cout << "Time spent pushing: " << push_time << endl; 
  }
  
  /***
  Primary interface to Quip. Requires no external setup other than 
  constructing the QuipSolver object.
  ***/ 
  ReturnValue QuipSolver::solve() {
    ReturnValue rv;
    {
      QuipAutoTimer auto_timer(stats.quip_time);
      rv = quipInit();
      bool trivial = true;
      
      if(rv.returnType == MC::Unknown) {
        rv = infBlockCube();
        trivial = false;
      }

      if(!trivial) {
        // For testing the support graph construction
        if (opts.do_final_sg) {
          constructSupportGraph();
        }

        if (opts.sg_type != NO_SG) {
          writeSupportGraph();
        }
      }
    }
      
    if (opts.print_stats){
      printStats();
    }

    return rv;
  }

  void QuipAction::exec()
  {
    QuipOptions opts(options());
    Options::Verbosity verbosity = model().verbosity();
    if (verbosity > Options::Silent){
      cout << "Starting Quip" << endl;
      if (verbosity >= Options::Verbose){
        //print all option settings
        cout << "verbosity " << verbosity << endl;
      }
    }
   
    QuipSolver quip(model(), opts); 
    ReturnValue rv = quip.solve();
    
    //set the result for iimc
    auto pat = model().attachment<ProofAttachment>(Key::PROOF);
    if (rv.returnType == MC::Proof) {
      pat->setConclusion(0);
      pat->setProof(rv.proof);
    }
    else if (rv.returnType == MC::CEX) {
      pat->setConclusion(1);
      pat->setCex(rv.cex);
    }
    model().release(pat);
  }
}
