/********************************************************************
Copyright (c) 2010-2015, Regents of the University of Colorado

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of the University of Colorado nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
********************************************************************/

#ifndef _QUIP_
#define _QUIP_

#include "ExprAttachment.h"
#include "CNFAttachment.h"
#include "ProofAttachment.h"
#include "RchAttachment.h"
#include "COI.h"
#include "Model.h"
#include "SAT.h"
#include "MC.h"
#include "options.h"
#include "IC3.h"

#include <boost/program_options.hpp>
#include <list>

namespace Quip {
  
  typedef std::vector<ID> Cube;
  typedef std::vector<ID> Clause;
  typedef uint64_t CubeID;

}
namespace std {

  template <>
  struct hash<Quip::Cube>
  {
    std::size_t operator()(const Quip::Cube& c) const;
  };

}


namespace Quip {

  typedef std::pair<const std::vector<ID> *, Expr::Manager::View *> PrintCubePair;

  enum ProofKind {MAY_PROOF, MUST_PROOF};

  void primeIDVector(Expr::Manager::View * ev,  Expr::IDVector & vec);

  class QuipLog {

    std::ostream * log_logorrheic;
    std::ostream * log_verbose;
    std::ostream * log_informative;
    std::ostream * log_terse;
    std::ostream * log_silent;

    // This code is straight from StackOverflow
    // (http://stackoverflow.com/a/7818394)
    class null_out_buf : public std::streambuf {
      public:
        virtual std::streamsize xsputn (const char *, std::streamsize n) {
            return n;
        }
        virtual int overflow (int) {
            return EOF;
        }
    };
    
    class null_out_stream : public std::ostream {
      public:
        null_out_stream() : std::ostream (&buf) {}
      private:
        null_out_buf buf;
    };

    null_out_stream null_stream;

    public:

      QuipLog() : 
          log_logorrheic(&null_stream),
          log_verbose(&null_stream),
          log_informative(&null_stream),
          log_terse(&null_stream),
          log_silent(&null_stream)
      { }

      QuipLog(Options::Verbosity verbosity) : 
          log_logorrheic(&null_stream),
          log_verbose(&null_stream),
          log_informative(&null_stream),
          log_terse(&null_stream),
          log_silent(&null_stream)
      {
        setVerbosity(verbosity);
      }

      void setVerbosity(Options::Verbosity verbosity) {
        log_logorrheic = &null_stream;
        log_verbose = &null_stream;
        log_informative = &null_stream;
        log_terse = &null_stream;
        log_silent = &null_stream;
        switch(verbosity) {
          case Options::Logorrheic:
            log_logorrheic = &std::cout;
          case Options::Verbose:
            log_verbose = &std::cout;
          case Options::Informative:
            log_informative = &std::cout;
          case Options::Terse:
            log_terse = &std::cout;
          case Options::Silent:
            log_silent = &std::cout;
            break;
          default:
            log_silent = &std::cout;
            break;
        }
      }

      std::ostream & logorrheic(bool c = true)   { return c ? *log_logorrheic  : null_stream; }
      std::ostream & verbose(bool c = true)      { return c ? *log_verbose     : null_stream; }
      std::ostream & informative(bool c = true)  { return c ? *log_informative : null_stream; }
      std::ostream & terse(bool c = true)        { return c ? *log_terse       : null_stream; } 
      std::ostream & silent(bool c = true)       { return c ? *log_silent      : null_stream; }
  };

  enum SupportGraphType { NO_SG                 = 0, 
                          ACTIVATION_LITERAL_SG = 1, 
                          UNSAT_CORE_SG         = 2,
                          FINAL_ONLY_SG         = 3, 
                          EXPERIMENTAL_SG       = 4,
                          HYBRID_SG             = 5,
                          LAST_SG_TYPE          = HYBRID_SG };

  struct QuipOptions {
    std::string backend;
    std::string sg_infdot_file;
    std::string sg_dot_file;
    int max_reachable_states;
    SupportGraphType sg_type;
    bool print_stats;
    bool renew_sat;
    bool print_cex;
    bool under_generalize;
    bool use_ctps;
    bool lift;
    bool abbr_backend;
    bool ugly_lemmas;
    bool do_final_sg;
    bool desperate_sg;
    bool reenqueue;
    bool sg_even_only;
    int min_sg_level;
    int sg_effort_limit;
    int sg_effort_limit_minlvl;
    // These are intended to support IICTL. They tend to just override the
    // related concepts from the model. For instance, iictl_clauses is a CNF
    // representation of the property, which normally comes from the model's
    // first ouput.
    // init_cube is a bit different, it should be the true initial state.
    // It's used to avoid promoting lemmas without initation relative to the
    // true initial state
    SAT::Clauses * iictl_clauses;
    SAT::Clauses * neg_constraints;
    std::vector<SAT::Clauses> * constraints;
    Cube * init_cube;

    QuipOptions() : backend("picosat"), 
                    sg_infdot_file(""),
                    sg_dot_file(""),
                    max_reachable_states(10000),
                    sg_type(NO_SG),
                    print_stats(false),
                    renew_sat(false),
                    print_cex(false),
                    under_generalize(true),
                    use_ctps(false),
                    lift(true),
                    abbr_backend(false),
                    ugly_lemmas(false),
                    do_final_sg(false),
                    desperate_sg(false),
                    reenqueue(true),
                    sg_even_only(false),
                    min_sg_level(1),
                    sg_effort_limit(INT_MAX),
                    sg_effort_limit_minlvl(INT_MAX),
                    iictl_clauses(NULL),
                    neg_constraints(NULL),
                    constraints(NULL),
                    init_cube(NULL)
    { }

    QuipOptions(
        const boost::program_options::variables_map & opts,
        const std::string * _backend = NULL) : QuipOptions()
    {
      backend = (_backend && _backend->empty()) ? *_backend : opts["quip_backend"].as<std::string>();
      max_reachable_states = opts["quip_max_reachable_states"].as<int>();
      sg_infdot_file = opts["quip_sg_infdot"].as<std::string>();
      sg_dot_file = opts["quip_sg_dot"].as<std::string>();
      print_stats = opts.count("quip_stats");
      renew_sat = opts.count("quip_renew_sat");
      print_cex = opts.count("print_cex");
      int sg_int = opts["quip_sg_type"].as<int>();
      sg_type = sg_int >= NO_SG && sg_int <= LAST_SG_TYPE ? (SupportGraphType) sg_int : NO_SG;
      use_ctps = opts.count("quip_use_ctps");
      lift = !opts.count("quip_xlift");
      abbr_backend = opts.count("quip_abbr_backend");
      ugly_lemmas = opts.count("quip_ugly");
      do_final_sg = opts.count("quip_final_sg") || sg_type == FINAL_ONLY_SG;
      desperate_sg = opts.count("quip_desperate_sg");
      min_sg_level = opts["quip_min_sg_level"].as<int>();
      reenqueue = !opts.count("quip_xreenq");
      sg_even_only = opts.count("quip_sg_even_only");
      sg_effort_limit = opts["quip_sg_effort_limit"].as<int>();
      sg_effort_limit_minlvl = opts["quip_sg_effort_limit_minlvl"].as<int>();

      // Only glucose supports abbreviations
      assert(!(abbr_backend && backend != "glucose"));

      // Without ugly lemmas, the effort limits cause non-termination
      assert(sg_effort_limit == INT_MAX || ugly_lemmas);
      assert(sg_effort_limit_minlvl == INT_MAX || ugly_lemmas);
    }

  };
  
  class QuipStats {
  public:
    QuipStats() : 
      level(0),
      init_states(0),
      sat_calls(0),
      consecution_calls(0),
      cons_from_prop_cls(0),
      syntactic_block_success(0),
      syntactic_block_fail(0),
      cube_size_total(0),
      cube_size_cnt(0),
      cube_reduction_total(0),
      cube_reduction_cnt(0),
      proof_size(0),
      core_size_total(0),
      core_vs_frame_size(0),
      core_cnt(0),
      num_ctps(0),
      ctp_cons_skipped(0),
      ctp_subsummption_checks(0),
      supp_cons_skipped(0),
      bad_lemmas(0),
      total_lemmas(0),
      may_fail(0),
      may_success(0),
      must_success(0),
      may_time(0),
      must_time(0),
      support_time(0),
      quip_time(0),
      push_time(0),
      ugly_lemmas(0),
      ugly_marked(0),
      ugly_unmarked(0),
      support_calls(0)
    {
    }
    void print() const;
    
    int level; 
    int init_states;
    int sat_calls;
    int consecution_calls; 
    int cons_from_prop_cls;
    int syntactic_block_success; //want to check how many sat calls actually avoided by doing this
    int syntactic_block_fail;

    //distribution of cubes (number of cubes in each frame at every level)
    //last element is size of infcubes
    std::vector< std::vector<int> > cube_dist_before_prop;               
    std::vector< std::vector<int> > cube_dist_after_prop;  
    //average number of clauses propagated in each call to propagateClauses
    std::vector<double> avg_cls_prop; 
    
    int cube_size_total;
    int cube_size_cnt;
    double cube_reduction_total; 
    int cube_reduction_cnt;
    int proof_size;
    int core_size_total;
    double core_vs_frame_size; //average of (unsat core size) / (size of F_k + size of F_inf)
    int core_cnt;
    int num_ctps;
    int ctp_cons_skipped; //number of consecution calls skipped due to CTP
    int ctp_subsummption_checks; 
    int supp_cons_skipped; //number of consecution calls skipped because of support lemmas
    int bad_lemmas;
    unsigned total_lemmas;
    
    //stats on proof obligations
    int may_fail;
    int may_success;
    int must_success;
    std::vector<int> may_proof_dist;
    std::vector<int> must_proof_dist;
    std::vector<int> effort_by_level;
    std::vector<int> effort_dist;

    double may_time, must_time, support_time, quip_time, push_time;
    unsigned ugly_lemmas, ugly_marked, ugly_unmarked;
    unsigned support_calls;
    
    void recordStat(std::vector<int> & stat,  int x);
    std::string str(const std::vector<int> & stat) const;
  };
  
    
  struct CTP { //counterexample to pushing
    Cube state;
    bool blocked;
    bool can_delete; //hacky way to avoid storing CTPs in a set
    CTP(const Cube & st) :  state(st), 
                            blocked(false),
                            can_delete(false)
    {}
  };

  class Lemma {
  public:
    Lemma () : bad(false), 
               ugly(false),
               is_init(false),
               ctp(NULL), 
               lv(-1),
               annotation("")
    { }

    Lemma (Cube c, int lv, bool bad=false) : bad(bad), 
                                             ugly(false),
                                             is_init(false),
                                             ctp(NULL),
                                             lv(lv),
                                             annotation(""),
                                             cube(c)
    { }

    const Cube & getCube() const { return cube; }

    bool bad;
    bool ugly;
    bool is_init;
    CTP * ctp;
    int lv;
    std::string annotation;
  private:
    Cube cube;
 
  };

  class CubeComp {
  public:
    bool operator()(const Cube & a, const Cube & b){
      return IC3::_IDVectorComp(a, b);
    }
  };
  typedef std::set<Cube, CubeComp> CubeSet;
  
  struct ReturnValue {
    ReturnValue(MC::ReturnType _returnType = MC::Unknown) : returnType(_returnType) {}
    MC::ReturnType returnType;
    std::vector<Transition> cex;
    std::vector<Cube> proof;
  };
    
  class ReachableState {
  public:
    ReachableState() {}
    ReachableState(Cube _state, Cube _inputs) :
      state(_state),
      pre(NULL),
      inputs(_inputs)
      {}
    Cube state;
    ReachableState * pre; //predecessor state 
    Cube inputs; //input assignment which led to this state
  };
  
  class ReachableStateManager {
  public:
    ReachableStateManager() :
      maxn_states(10000),
      insertion_policy(0.5),
      total_states(0),
      num_intersections(0),
      num_evictions(0)
      {}
      
    /***
    maxn_states: maximum number of reachable states to keep
    insertion_policy: double between 0 and 1 defining where in the list new states are 
    inserted. 0 = front of the list, 1 = end of the list, 0.5 = middle of the list, etc.
    ***/
    ReachableStateManager(int _maxn_states, double _insertion_policy=0.5) : 
      total_states(0),
      num_intersections(0),
      num_evictions(0)
    {
      maxn_states = _maxn_states;
      insertion_policy = _insertion_policy;
    }
    
    ~ReachableStateManager(){
      clear();
    }
    
    ReachableState * addState(const Cube & cube, const Cube & inputs);
    ReachableState * intersect(const Cube & cube);
    void printStats() const;
    int getNumStates() const { return states.size(); }
    void clear() {
      for (auto it=states.begin(); it!=states.end(); it++){
        delete (*it);
      }
      for (auto it=evicted_states.begin(); it!=evicted_states.end(); it++){
        delete (*it);
      }
      states.clear();      
    }
    
  private:
    std::list<ReachableState*> states;
    std::list<ReachableState*> evicted_states; //no longer active states 
        //(keep around for building cex trace)
    int maxn_states;
    double insertion_policy;
    //stats
    int total_states;           
    int num_intersections; 
    int num_evictions;    
  };
  
  struct ProofObligation {
    ProofObligation(const Cube & _cube, 
                    int _level, 
                    ProofKind _type, 
                    ProofObligation * _parent = NULL,
                    int _effort_limit = INT_MAX) : 
                    cube(_cube), 
                    level(_level), 
                    type(_type), 
                    parent(_parent), 
                    from_sg(false), 
                    sg_effort(0)
    {
      if(_parent) {
        effort_limit =  _parent->effort_limit == INT_MAX ? INT_MAX : 
                        _parent->effort_limit - _parent->effort;
        depth = _parent->depth + 1;
      } 
      else {
        effort_limit = _effort_limit;
        depth = 0;
      }
      effort = 0; 
      failed = false;
      active = true;
    }

    ProofObligation(const ProofObligation & other) 
    {
      cube      = other.cube;
      level     = other.level;
      type      = other.type;
      parent    = other.parent;
      effort_limit = other.effort_limit;
      depth     = other.depth;
      effort    = other.effort; 
      failed    = other.failed;
      active    = true;
      cex_state = other.cex_state;
      inputs    = other.inputs;
      from_sg   = other.from_sg;
      sg_effort = other.sg_effort;
    }
 
    Cube cube;
    Cube cex_state;
    Cube inputs; //for reconstructing the cex trace
    int level;
    ProofKind type; //may or must
    int effort; 
    int effort_limit;
    ProofObligation * parent;
    int depth;
    bool failed;
    bool active; //whether or not it's currently in the queue
    bool from_sg;
    int sg_effort;
  };

  class ProofObligationComparator {
  public:
    bool operator()(const ProofObligation * a, const ProofObligation * b) {
      if (a->type == b->type) {
        if (a->level == b->level) {
          return a->cube.size() > b->cube.size();
        }
        return a->level > b->level; //"largest" element for priority queue is lowest level
      } else {
        assert(a->type == MAY_PROOF || a->type == MUST_PROOF);
        assert(b->type == MAY_PROOF || b->type == MUST_PROOF);
        return a->type == MUST_PROOF;
      }
    }
  };

  class CubeDatabase;
  class SupportGraphDotWriter;
  class SupportGraph;
  class ConsecutionManager;

  class QuipSolver {
    enum ConsecutionType {
      CONSECUTION_NO_SG, 
      CONSECUTION_SG, 
      CONSECUTION_ACCEPTABLE_SG, 
      LAST_CONSECUTION_TYPE = CONSECUTION_ACCEPTABLE_SG
    };
    typedef std::set<Cube> CubeSet;
  public:
    QuipSolver(Model & m, QuipOptions o); 
    QuipSolver(Model & m) : QuipSolver(m, QuipOptions()) { } 
    ~QuipSolver();

    // The primary interface to Quip
    ReturnValue solve();
    // Interface for separate initialization/block steps 
    // (perhaps an incremental setting)
    ReturnValue quipInit();
    ReturnValue infBlockCube();
    ReturnValue infBlockCube(const Cube & bad_cube);
    ReturnValue blockCube(const Cube & bad_cube, int k, int effort_limit=INT_MAX);
    
    // Incremental interface to reset the parts that need reseting,
    // optionally with new QuipOptions
    void prepare(QuipOptions * new_opts = NULL, bool clear_finf = true, bool rebuild_trace = false);
    void cleanup();

    void constructSupportGraph(int level);
    void writeSupportGraph() const;
    void printStats();
    
  private:
    const Model & model() const { return m; }
    const QuipOptions & options() const { return opts; } 

    //helper functions
  public:
    static bool subsumes(const Cube & a, const Cube & b);
  private:
    // Use an overload instead of a default value for sep so it can be called
    // from within gdb more easily.
    std::string stringOfClauseOrCube(const std::vector<ID> & clause_or_cube) const {
      return stringOfClauseOrCube(clause_or_cube, "");
    }
    
    std::string stringOfClauseOrCube(const std::vector<ID> & clause_or_cube, const std::string & sep) const;
    std::string scube(CubeID id) const;

    PrintCubePair pc(const Cube & cube) const {
      return std::make_pair(&cube, ev);
    }
    
    ID getNPI() const;

    void logObligation(const ProofObligation * po) const;
    ReachableState * handleFailedMayProof(ProofObligation * po, const Cube & pred);
    ID levelVar(int k) const;
    ID newSGVar(const Cube & cube);
    ID newSGVar(CubeID id);
    const Cube & cubeFromSGVar(ID lit) const;
    ID SGVar(const Cube & cube);
    bool isSGVar(ID lit) const;
    Lemma & getLemma(const Cube & cube);
    Lemma & getLemma(CubeID id);
    const Lemma & getLemma(const Cube & cube) const;
    const Lemma & getLemma(CubeID id) const;
    void markBad(const Cube & cube);
    void markUgly(const Cube & cube);
    void unmarkUgly(const Cube & cube);
    bool isKnownLemma(const Cube & cube) const;
    bool isAcceptable(const Cube & cube) const;
    bool isAcceptable(CubeID id) const;
    bool hasAcceptableSupport(const Cube & cube) const;
    bool hasBadSupport(const Cube & cube) const;
    ID var(ID lit) const;
    void logAddCube(const std::string & prefix, int lv, const Cube & cube, std::ostream & stream) const;
    void checkCTPs(const Cube & cube, int lv, int min_check=1);
    void addCubeToManager(const Cube & cube, int lv, SAT::Manager * manager = NULL);
    void addCubeToSGManager(const Cube & cube, bool act, SAT::Manager * manager = NULL);
    void addCubeSetToManager(const CubeSet & cubes, int lv, SAT::Manager * manager = NULL);
    void addCubeSetToSGManager(const CubeSet & cubes, bool act, SAT::Manager * manager = NULL);
    int addCubeToTrace(const Cube & cube, int lv);
    void addCube(const Cube & cube, int lv);
    Cube getSGAssumps(int lvl, bool acceptable_only = false);
    void getCubesFromAsgn(const SAT::Assignment & asgn, 
                                Cube * cube = NULL, 
                                Cube * input_cube = NULL, 
                                Cube * p_input_cube = NULL,
                                Cube * p_state_vars = NULL) const;

    bool consecution(const Cube & c, int k, 
                           SAT::Assignment * asgn = NULL, 
                           Cube * reduced = NULL, 
                           bool do_support_graph = false);
    void computeReduced(const Cube & c, Cube & reduced, const Cube & crits) const;
    bool consecutionStandard(const Cube & c, int k, SAT::Assignment * asgn, Cube * reduced, ConsecutionType type);
    bool consecutionActivation(const Cube & c, int k, SAT::Assignment * asgn, Cube * reduced, ConsecutionType type);
    void computeSupportSetFromCrits(const Cube & crits, std::set<Cube> & support) const;
    int getSupportLevel(const Cube & cube) const;
    bool computeAcceptableSupport(const Cube & cube, int lvl = -1);
    void computeSupport(const Cube & cube);
    void computeSupport(const Cube & cube, int lvl);

    ReachableState * extractReachable(ProofObligation * po, const Cube & r0);
    void initTR(); 
    void getTRClauses(std::vector<Clause> * cons_clauses = NULL, 
                      std::vector<Clause> * error_clauses = NULL, 
                      ID npi = 0, bool silent = false) const;
    ReturnValue buildCex(ProofObligation & obl, ReachableState * r0 = NULL) const;
    void printReachableStates() const;
    int getLevel() const { return current_level; }
    void setLevel(int lv) { 
      assert(lv >= 0); 
      current_level = lv; 
    }

    void incrementLevel() { 
      current_level++; 
    }

    //functions for generalization (adapted from IC3)
    bool initiation(const Cube & cube) const;
    bool join(Cube & cube, const SAT::Assignment & asgn, const std::set<ID> & keep) const;
    bool down(Cube & cube, int k, const std::set<ID> & keep);
    void mic(Cube & cube, int k);
    int generalize(int lo, int hi, Cube & cube);
    void unsatCoreLift(const Cube & t, const Cube & inputs, const Cube & p_inputs, const Cube & s, Cube * lifted = NULL) const;
    void initializeAsgn(SAT::Assignment & asgn, bool primed_state = false) const;

    bool checkInitial(const Cube & bad_cube) const;
    std::pair<Cube, int> block(const ProofObligation & po, Cube * inputs = NULL, Cube * concrete_state = NULL);
    MC::ReturnType propagateClauses();
    
    void constructSupportGraph();
    void addCoreToGraph(const std::vector<Clause> & core, const Cube & c, int k);

    // (re-)initialization related functions
    Cube getCurrentBad() const;
    void initFrames();
    void initLiftingSAT();
    void initSAT();
    void initSGSAT();
    bool loadTR(SAT::Manager * manager);
    void getIICTLClauses(std::vector<Clause> & cons_clauses, std::vector<Clause> & error_clauses, ID npi);
    void loadFrames(SAT::Manager * manager = NULL);
    void loadFramesWithSGLits(SAT::Manager * manager, bool inf_act = false);
    SAT::Manager * initSATManager(bool load_frames = true, bool load_inf = true);
    SAT::Manager * initSATManagerWithoutFrames() {
      return initSATManager(false, false);
    }
    bool renewSAT(bool sg_man_inf = false);
    void cleanupSAT();
    void rebuildTrace();
    void growFrames(size_t k);
    void cleanupCTPs();
    void cleanupTR();

    Model & m;
    Options::Verbosity verbosity;
    QuipOptions opts;
    mutable QuipLog logger;
    mutable QuipStats stats;

    Expr::Manager::View * ev;
    // For consecution
    SAT::Manager * sat_man;
    // For lifting
    SAT::Manager * lift_man;
    // For computing support graphs
    SAT::Manager * sg_man;
    SAT::Manager::View * sat_view;
    SAT::Manager::View * lift; 
    SAT::Manager::View * sg_view; 

    std::vector<Clause> tr_clauses;
    std::vector<CubeSet> frames;
    CubeSet inf_cubes;
    std::vector<ID> state_vars;
    std::vector<ID> input_vars;
    Cube init_cube;
    std::set<ID> initially;
    CubeDatabase * cube_db;
    SupportGraph * support_graph;
    ReachableStateManager rsman;
    std::set<ID> coi_latches;
    int current_level;

    std::unordered_map<CubeID, ID> cube_to_sg_lit;
    std::unordered_map<ID, CubeID> sg_lit_to_cube;
    size_t next_sg_lit_index;
    bool sg_man_has_inf_act;

    std::vector< std::list<CTP*> > active_ctps; //active CTPs for each level
  };
  
  class QuipAction : public Model::Action {
  public:
    QuipAction(Model & m) : Model::Action(m)
    {
      ExprAttachment::Factory eaf;
      requires(Key::EXPR, &eaf);
      COIAttachment::Factory caf;
      requires(Key::COI, &caf);
      RchAttachment::Factory raf;
      requires(Key::RCH, &raf);
      ProofAttachment::Factory paf;
      requires(Key::PROOF, &paf);
      CNFAttachment::Factory cnfaf;
      requires(Key::CNF, &cnfaf);
    }

    virtual void exec();

    
  private:
    static ActionRegistrar action;
  };
  
}

#endif
